#!/bin/bash -xv

# script to estimate absolute photometric
# calibration on a night basis.

# $1 main dir
# $2 standard dir
# $3 science dir
# $4 image extension 

INSTDIR=instruments_professional
if [ ! -f ${INSTDIR}/${INSTRUMENT}.ini ]; then
    INSTDIR=instruments_commercial
fi
if [ ! -f ${INSTDIR}/${INSTRUMENT}.ini ]; then
    INSTDIR=~/.theli/instruments_user/
fi
. ${INSTDIR}/${INSTRUMENT:?}.ini

. bash_functions.include
. bash.include

if [ ! -d "/$1/$2/calib" ]; then
  mkdir /$1/$2/calib
fi

# find the chip catalogs that should be used for photometry,
# i.e. reject chips that have the NOTUSE flag set.

CATS=""
i=1
while [ ${i} -le ${NCHIPS} ]
do
  if [ ${NOTUSE[${i}]:=0} -eq 0 ] && [ ${NOTPROCESS[${i}]:=0} -eq 0 ] && [ -f /$1/$2/cat/chip_${i}_merg.cat ]; then
    CATS="${CATS}  /$1/$2/cat/chip_${i}_merg.cat"
  fi
  i=$(( $i + 1 ))
done

if [ "${CATS}_A" = "_A" ]; then
    exit
fi

${P_LDACPASTE} -i ${CATS} -t PSSC -o ${TEMPDIR}/tmp_chips_$$.cat

${P_LDACFILTER} -i ${TEMPDIR}/tmp_chips_$$.cat -t PSSC \
    -c "((((${V_ABSPHOT_STDFILTER}mag < 99)AND(${V_ABSPHOT_STDCOLOR} < 99))AND(Mag < 0))AND(${V_ABSPHOT_STDFILTER}mag > ${V_ABSPHOT_STDMINMAG}));" \
    -o /$1/$2/cat/allchips_merg.cat

${P_LDACFILTER} -i /$1/$2/cat/allchips_merg.cat -t PSSC \
    -c "(${V_ABSPHOT_STDCOLOR} > -10) AND (${V_ABSPHOT_STDCOLOR} < 10);" \
    -o /$1/$2/cat/allchips_tmp.cat

LABEL=`echo ${V_ABSPHOT_STDCOLOR} | sed 's/m/-/g'`

# Get a list of all nights and write it to the FIFO
NIGHTS=`${P_LDACTOASC} -i /$1/$2/cat/allchips_tmp.cat  -t PSSC \
    -b -k GABODSID | ${P_SORT} | uniq | ${P_GAWK} '{printf("%s ", $1)}'`

for NIGHT in ${NIGHTS}
do
  echo 
  echo "       ---==== Calibrating night ${NIGHT} ====---"
  echo
  ${P_LDACFILTER} -i /$1/$2/cat/allchips_tmp.cat  -t PSSC \
      -o /$1/$2/cat/night_${NIGHT}.cat -c "(GABODSID=${NIGHT});"
  ${P_LDACTOASC} -i /$1/$2/cat/night_${NIGHT}.cat  -t PSSC \
      -b -k Mag ${V_ABSPHOT_STDFILTER}mag ${V_ABSPHOT_STDCOLOR} AIRMASS > \
      ${TEMPDIR}/night_${V_ABSPHOT_FILTER}_${NIGHT}_$$.asc

  ${S_PHOTOABS} --input=${TEMPDIR}/night_${V_ABSPHOT_FILTER}_${NIGHT}_$$.asc \
      --output=${TEMPDIR}/photo_res \
      --extinction="${V_ABSPHOT_EXTINCTION}" \
      --color="${V_ABSPHOT_COLORFIXED}" \
      --night=${NIGHT} \
      --label=${LABEL} \
      --batch

  # in 'photo_abs.py' we 'abuse' the return value to give the interactive
  # choice back to the shell:
  CHOICE=$?
  mv ${TEMPDIR}/photo_res.asc \
      /$1/$2/calib/night_${NIGHT}_${V_ABSPHOT_FILTER}_result.asc
  mv ${TEMPDIR}/photo_res.png \
      /$1/$2/calib/night_${NIGHT}_${V_ABSPHOT_FILTER}_result.png

  ZP[0]=-1.0
  COEFF[0]=-1.0
  
  i=1
  while read -a SOL
  do
      ZP[$i]=${SOL[0]}
      COEFF[$i]=${SOL[1]}
      COL[$i]=${SOL[2]}
      i=$(($i + 1))
  done < /$1/$2/calib/night_${NIGHT}_${V_ABSPHOT_FILTER}_result.asc
  
  CHOICE=`${P_GAWK} 'BEGIN{if('${ZP[2]}' > '${V_ABSPHOT_ZPMIN}' && '${ZP[2]}' < '${V_ABSPHOT_ZPMAX}') {
            print 2;
        } 
        else 
        {
            print 0;
            }}'`

  ZPCHOICE=${CHOICE}

  echo
  echo "Updating image headers ..."
  echo

  # for those images that DO have a STD observed in the same night

  ${P_DFITS} /$1/$3/*$4.fits | ${P_FITSORT} -d GABODSID | \
      ${P_GAWK} '{if ($2 == '${NIGHT}') print $1}' > \
      ${TEMPDIR}/night_${NIGHT}_img_list_$$.asc
  
  while read IMG
  do
    echo $IMG
    # Write the choice to the header
    value ${ZPCHOICE}
    writekey /$1/$3/${IMG} ZPCHOICE "${VALUE}" REPLACE

    value ${ZP[$CHOICE]}
    writekey /$1/$3/${IMG} ZP "${VALUE}" REPLACE

    value ${COEFF[$CHOICE]}
    writekey /$1/$3/${IMG} COEFF "${VALUE}" REPLACE

    i=1
    while [ "${i}" -le "3" ]
    do
	value ${ZP[$i]}
	writekey /$1/$3/${IMG} ZP${i} "${VALUE}" REPLACE
	
	value ${COEFF[$i]}
	writekey /$1/$3/${IMG} COEFF${i} "${VALUE}" REPLACE
	
	i=$(( $i + 1 ))
    done
  done < ${TEMPDIR}/night_${NIGHT}_img_list_$$.asc
done


# for those images that DO NOT have a STD observed in the same night,
# we need the ZPCHOICE (= 0) keyword for consistency

images=`${P_FIND} /$1/$3/ -maxdepth 1 -name \*$4.fits`

for image in ${images}
do
    zpc=`get_key ${image} ZPCHOICE`
    if [ "${zpc}"_A = "_A" ]; then
	value 0
	writekey ${image} ZPCHOICE "${VALUE} / No STD observed this night" REPLACE
	i=1
	while [ "${i}" -le "3" ]
	do
	    value -1.0
	    writekey ${image} ZP${i} "${VALUE} / No STD observed this night" REPLACE
	    
	    value -1.0
	    writekey ${image} COEFF${i} "${VALUE} / No STD observed this night" REPLACE
	    
	    i=$(( $i + 1 ))
	done
    fi
done

rm ${TEMPDIR}/night_${V_ABSPHOT_FILTER}_*_$$.asc \
   ${TEMPDIR}/night_*_img_list_$$.asc \
   ${TEMPDIR}/tmp_chips_$$.cat
