#!/bin/bash -xv

# this script retrieves and reformats a 
# standard star catalogue from the CDS server

# $1: main dir
# $2: science dir
# $3: image extension (OFC etc.)
# $4: the catalogue queried
# $5: server

. progs.ini

maindir=$1/$2/
imext=$3
catid=$4
server=$5
#server=${V_AP_SERVER}
ra=${V_AP_REFRA}
dec=${V_AP_REFDEC}
size=${V_AP_RADIUS}
maglim=${V_AP_MAGLIM}

echo $catid > ${maindir}/.astrefcat

dir=`pwd`

# if the coordinates are determined from the images

if [ "$1" != "_" ] && [ "$2" != "_" ]; then 
    if [ "${ra}" = "header" ] || [ "${dec}" = "header" ]; then
	cd ${maindir}
	ls *${imext}.fits > ${TEMPDIR}/imcoord.list_$$
	if [ -f ${TEMPDIR}/imcoord.coords_$$ ]; then
	    rm ${TEMPDIR}/imcoord.coords_$$
	fi
	
        # obtain the sky coordinates for all images in the set.
	cat ${TEMPDIR}/imcoord.list_$$ | \
	    {
	    while read file
	    do
		crval1=`${P_DFITS} ${file} | ${P_FITSORT} CRVAL1 |\
	    ${P_GAWK} '(NR!=1) {printf "%4.6f\n", $2}'`
		crval2=`${P_DFITS} ${file} | ${P_FITSORT} CRVAL2 |\
	    ${P_GAWK} '(NR!=1) {printf "%4.6f\n", $2}'`
		echo ${crval1} ${crval2} >> ${TEMPDIR}/imcoord.coords_$$
	    done
	}
	
        # calculate the median position and the extent in alpha and delta;
        # using the median instead of the mean we avoid problems with the 
        # discontinuous jump when crossing from RA=0 to RA=360 degrees.
	${P_GAWK} '{print $1}' ${TEMPDIR}/imcoord.coords_$$ > ${TEMPDIR}/alpha_$$
	${P_GAWK} '{print $2}' ${TEMPDIR}/imcoord.coords_$$ > ${TEMPDIR}/delta_$$
	
	ra=`${P_GETSTATS} -i ${TEMPDIR}/alpha_$$ -m median -n -f`
	delta=`${P_GETSTATS} -i ${TEMPDIR}/delta_$$ -m median -f`
	
	dec=`echo $delta | ${P_GAWK} '{if ($1>0) print "+"$1; else print $1}'`
	
	cd $dir
    fi
fi

# the CDS client wants an explicit "+" if dec is positive

dec=`echo ${dec} | ${P_GAWK} 'BEGIN {FS=":"} {if ($1>0 && $1!~/+/) print "+"$0; else print $0}'`

VIZOPTION="-wget -mime=csv -site=${server} -out.max=100000000 -c.rm=${size} -out=_RAJ -out=_DEJ"

#obtain data from the allsky catalog
if [ "${catid}" = "ALLSKY" ]; then
   ${P_GAWK} '($3<'${maglim}')' ../astrstdcats/gsc1_allsky.cat > ${TEMPDIR}/tmp1_$$
fi

# obtain data from the GSC1
if [ "${catid}" = "GSC-1.2" ]; then
    ${P_VIZQUERY} ${VIZOPTION} -c="$ra $dec" -source=GSC1.2 \
	Pmag=0..${maglim} \
	-out=Pmag | \
	${P_GAWK} '($0 !~/#/)' | ${P_GAWK} '(NR>4&&NF>0)' | sed 's/;/ /g' \
	> ${TEMPDIR}/tmp1_$$
fi

# obtain data from the PPMXL
if [ "${catid}" = "PPMXL" ]; then
    ${P_VIZQUERY} ${VIZOPTION} -c="$ra $dec" -source=I/317 \
	-out=b2mag -out=r2mag -out=imag | \
	${P_GAWK} '($0 !~/#/)' | ${P_GAWK} '(NR>4&&NF>0)' | \
	${P_GAWK} 'BEGIN{FS=";"} {
             av=0
             sum=0
             if ($3 > 0.) {av+=$3; sum++}
             if ($4 > 0.) {av+=$4; sum++}
             if ($5 > 0.) {av+=$5; sum++}
             if (sum !=0) {printf ("%s %s %f\n", $1, $2, av/sum)}
             }' > ${TEMPDIR}/tmp1_$$
fi

#obtain data from TYC
if [ "${catid}" = "TYC" ]; then
    ${P_VIZQUERY} ${VIZOPTION} -c="$ra $dec" -source=TYC/tyc_main \
	Vmag=0..${maglim} \
	-out=Vmag | \
	${P_GAWK} '($0 !~/#/)' | ${P_GAWK} '(NR>4&&NF>0)' | sed 's/;/ /g' \
	> ${TEMPDIR}/tmp1_$$
fi

# obtain data from the GSC2
if [ "${catid}" = "GSC-2.3" ]; then
    ${P_VIZQUERY} ${VIZOPTION} -c="$ra $dec" -source=GSC2.3 \
	-out="Fmag jmag Vmag Nmag" | \
	${P_GAWK} '($0 !~/#/)' | ${P_GAWK} '(NR>4&&NF>0)' | \
	${P_GAWK} 'BEGIN{FS=";"} {
             av=0
             sum=0
             if ($3 > 0.) {av+=$3; sum++}
             if ($4 > 0.) {av+=$4; sum++}
             if ($5 > 0.) {av+=$5; sum++}
             if ($6 > 0.) {av+=$6; sum++}
             if (sum !=0) {printf ("%s %s %f\n", $1, $2, av/sum)}
             }' > ${TEMPDIR}/tmp1_$$
fi

# obtain data from UCAC3
if [ "${catid}" = "UCAC3" ]; then
    ${P_VIZQUERY} ${VIZOPTION} -c="$ra $dec" -source=UCAC3 \
	-out="f.mag" f.mag=0..${maglim} | \
	${P_GAWK} '($0 !~/#/)' | ${P_GAWK} '(NR>4&&NF>0)' | \
	${P_GAWK} 'BEGIN{FS=";"} {
             printf ("%s %s %f\n", $1, $2, $3)
             }' > ${TEMPDIR}/tmp1_$$
fi


# obtain data from the USNO-B1
if [ "${catid}" = "USNO-B1" ]; then
    ${P_VIZQUERY} ${VIZOPTION} -c="${ra} ${dec}" -source=USNOB1 \
	-out="B1mag B2mag R1mag R2mag Imag" | \
	${P_GAWK} '($0 !~/#/)' | ${P_GAWK} '(NR>4&&NF>0)' | \
	${P_GAWK} 'BEGIN{FS=";"} {
             av=0
             sum=0
             if ($3 > 0.) {av+=$3; sum++}
             if ($4 > 0.) {av+=$4; sum++}
             if ($5 > 0.) {av+=$5; sum++}
             if ($6 > 0.) {av+=$6; sum++}
             if ($7 > 0.) {av+=$7; sum++}
             if (sum !=0) {printf ("%s %s %f\n", $1, $2, av/sum)}
             }' > ${TEMPDIR}/tmp1_$$
fi

# obtain data from the 2MASS
if [ "${catid}" = "2MASS" ]; then
    ${P_VIZQUERY} ${VIZOPTION} -c="$ra $dec" -source=2MASS-PSC \
	-out="Jmag Hmag Kmag" | \
	${P_GAWK} '($0 !~/#/)' | ${P_GAWK} '(NR>4&&NF>0)' | \
	${P_GAWK} 'BEGIN{FS=";"} {
            av=0
            sum=0
            if ($3 > 0.) {av+=$3; sum++}
            if ($4 > 0.) {av+=$4; sum++}
            if ($5 > 0.) {av+=$5; sum++}
            if (sum !=0) {printf ("%s %s %f\n", $1, $2, av/sum)}
            }' > ${TEMPDIR}/tmp1_$$
fi

# obtain data from the SDSS-DR7
#if [ "${catid}" = "SDSS-DR7" ]; then
#    ${P_VIZQUERY} ${VIZOPTION} -c="$ra $dec" -source=SDSS-DR7 \
#	-out="gmag rmag imag" | \
#	${P_GAWK} '($0 !~/#/)' | ${P_GAWK} '(NR>4&&NF>0)' | \
#	${P_GAWK} 'BEGIN{FS=";"} {
#          av=0
#         sum=0
#         if ($3 > 0.) {av+=$3; sum++}
#        if ($4 > 0.) {av+=$4; sum++}
#       if ($5 > 0.) {av+=$5; sum++}
#      if (sum !=0) {printf ("%s %s %f\n", $1, $2, av/sum)}
#     }' > ${TEMPDIR}/tmp1_$$
#fi

# obtain data from the SDSS-DR8
if [ "${catid}" = "SDSS-DR8" ]; then
    ramin=`echo ${ra} ${size} ${dec} | \
	${P_GAWK} '{ramin=$1-$2/60./cos($3*3.14159/180.); if (ramin<0) {ramin=ramin+360.}} END {print ramin}'`
    ramax=`echo ${ra} ${size} ${dec} | \
	${P_GAWK} '{ramax=$1+$2/60./cos($3*3.14159/180.); if (ramax>360) {ramax=ramax-360.}} END {print ramax}'`
    decmin=`echo ${dec} ${size} | ${P_GAWK} '{decmin=$1-$2/60.; if (decmin<-90) {decmin=-90}} END {print decmin}'`
    decmax=`echo ${dec} ${size} | ${P_GAWK} '{decmax=$1+$2/60.; if (decmax>90) {decmax=90}} END {print decmax}'`

    python ${dir}/sqlcl.py -l -f csv -q "SELECT ra,dec,r FROM PhotoObj WHERE ra BETWEEN ${ramin} AND ${ramax} AND dec BETWEEN ${decmin} AND ${decmax} AND r < ${maglim} AND r > 10 AND clean = 1" | sed 's/,/ /g' > ${TEMPDIR}/tmp1_$$
fi

# filter the magnitudes

if [ "${maglim}_A" != "_A" ]; then
    ${P_GAWK} '{if ($3<'${maglim}') printf ("%s %s %f\n", $1, $2, $3)}' \
	${TEMPDIR}/tmp1_$$ > ${TEMPDIR}/tmp1filt_$$
    mv ${TEMPDIR}/tmp1filt_$$ ${TEMPDIR}/tmp1_$$
fi

# EXIT if no sources left

numstars=`wc ${TEMPDIR}/tmp1_$$ | ${P_GAWK} '{print $1}'`

if [ "${numstars}" = "0" ]; then
    test -f ~/.theli/scripts/theli_mystd.skycat && rm ~/.theli/scripts/theli_mystd.skycat
    touch ~/.theli/scripts/theli_mystd.skycat
    exit 1
fi

# make an astrometrix catalogue

${P_DECSEXCONV} -i ${TEMPDIR}/tmp1_$$ -o ~/.theli/scripts/theli_mystd_astrometrix.cat

# make a scamp catalogue

echo VERBOSE = DEBUG         >  ${TEMPDIR}/makescamp.conf_$$
echo COL_NAME  = X_WORLD     >> ${TEMPDIR}/makescamp.conf_$$
echo COL_TTYPE = DOUBLE      >> ${TEMPDIR}/makescamp.conf_$$
echo COL_HTYPE = FLOAT       >> ${TEMPDIR}/makescamp.conf_$$
echo COL_COMM = '""'         >> ${TEMPDIR}/makescamp.conf_$$
echo COL_UNIT = '""'         >> ${TEMPDIR}/makescamp.conf_$$
echo COL_DEPTH = 1           >> ${TEMPDIR}/makescamp.conf_$$
echo COL_NAME  = Y_WORLD     >> ${TEMPDIR}/makescamp.conf_$$
echo COL_TTYPE = DOUBLE      >> ${TEMPDIR}/makescamp.conf_$$
echo COL_HTYPE = FLOAT       >> ${TEMPDIR}/makescamp.conf_$$
echo COL_COMM = '""'         >> ${TEMPDIR}/makescamp.conf_$$
echo COL_UNIT = '""'         >> ${TEMPDIR}/makescamp.conf_$$
echo COL_DEPTH = 1           >> ${TEMPDIR}/makescamp.conf_$$
echo COL_NAME  = MAG         >> ${TEMPDIR}/makescamp.conf_$$
echo COL_TTYPE = FLOAT       >> ${TEMPDIR}/makescamp.conf_$$
echo COL_HTYPE = FLOAT       >> ${TEMPDIR}/makescamp.conf_$$
echo COL_COMM = '""'         >> ${TEMPDIR}/makescamp.conf_$$
echo COL_UNIT = '""'         >> ${TEMPDIR}/makescamp.conf_$$
echo COL_DEPTH = 1           >> ${TEMPDIR}/makescamp.conf_$$
echo COL_NAME  = MAGERR      >> ${TEMPDIR}/makescamp.conf_$$
echo COL_TTYPE = FLOAT       >> ${TEMPDIR}/makescamp.conf_$$
echo COL_HTYPE = FLOAT       >> ${TEMPDIR}/makescamp.conf_$$
echo COL_COMM = '""'         >> ${TEMPDIR}/makescamp.conf_$$
echo COL_UNIT = '""'         >> ${TEMPDIR}/makescamp.conf_$$
echo COL_DEPTH = 1           >> ${TEMPDIR}/makescamp.conf_$$
echo COL_NAME  = ERRA_WORLD  >> ${TEMPDIR}/makescamp.conf_$$
echo COL_TTYPE = FLOAT       >> ${TEMPDIR}/makescamp.conf_$$
echo COL_HTYPE = FLOAT       >> ${TEMPDIR}/makescamp.conf_$$
echo COL_COMM = '""'         >> ${TEMPDIR}/makescamp.conf_$$
echo COL_UNIT = '""'         >> ${TEMPDIR}/makescamp.conf_$$
echo COL_DEPTH = 1           >> ${TEMPDIR}/makescamp.conf_$$
echo COL_NAME  = ERRB_WORLD  >> ${TEMPDIR}/makescamp.conf_$$
echo COL_TTYPE = FLOAT       >> ${TEMPDIR}/makescamp.conf_$$
echo COL_HTYPE = FLOAT       >> ${TEMPDIR}/makescamp.conf_$$
echo COL_COMM = '""'         >> ${TEMPDIR}/makescamp.conf_$$
echo COL_UNIT = '""'         >> ${TEMPDIR}/makescamp.conf_$$
echo COL_DEPTH = 1           >> ${TEMPDIR}/makescamp.conf_$$
echo COL_NAME  = ERRTHETA_WORLD  >> ${TEMPDIR}/makescamp.conf_$$
echo COL_TTYPE = FLOAT       >> ${TEMPDIR}/makescamp.conf_$$
echo COL_HTYPE = FLOAT       >> ${TEMPDIR}/makescamp.conf_$$
echo COL_COMM = '""'         >> ${TEMPDIR}/makescamp.conf_$$
echo COL_UNIT = '""'         >> ${TEMPDIR}/makescamp.conf_$$
echo COL_DEPTH = 1           >> ${TEMPDIR}/makescamp.conf_$$

${P_GAWK} '{print $0, "0.1", "0.00003", "0.00003", "0.1"}' \
    ${TEMPDIR}/tmp1_$$ > ${TEMPDIR}/tmp_scamp_$$

${P_ASCTOLDAC} -a ${TEMPDIR}/tmp_scamp_$$ \
               -o ${TEMPDIR}/tmp2_scamp0_$$ \
               -t OBJECTS -c ${TEMPDIR}/makescamp.conf_$$

${P_LDACRENTAB} -i ${TEMPDIR}/tmp2_scamp0_$$ \
    -o ${TEMPDIR}/tmp2_scamp1_$$ -t OBJECTS LDAC_OBJECTS 
${P_LDACRENTAB} -i ${TEMPDIR}/tmp2_scamp1_$$ \
    -o ~/.theli/scripts/theli_mystd_scamp.cat -t FIELDS LDAC_IMHEAD


# make a ds9 catalogue
echo global color=red select=0 highlite=1 wcs=wcs > \
    ~/.theli/scripts/theli_mystd.reg

${P_GAWK} '{print "j2000;circle", $1, $2, "5i"}' ${TEMPDIR}/tmp1_$$ >> \
    ~/.theli/scripts/theli_mystd.reg


# make a skycat catalogue
${P_GAWK} '{print NR, $1, $2, $3}' ${TEMPDIR}/tmp1_$$ > ${TEMPDIR}/skc_tmp2_$$

# make a skycat config file
echo VERBOSE = DEBUG         >  ${TEMPDIR}/makeskycat.conf_$$
echo COL_NAME  = SeqNr       >> ${TEMPDIR}/makeskycat.conf_$$
echo COL_TTYPE = LONG        >> ${TEMPDIR}/makeskycat.conf_$$
echo COL_HTYPE = INT         >> ${TEMPDIR}/makeskycat.conf_$$
echo COL_COMM = '""'         >> ${TEMPDIR}/makeskycat.conf_$$
echo COL_UNIT = '""'         >> ${TEMPDIR}/makeskycat.conf_$$
echo COL_DEPTH = 1           >> ${TEMPDIR}/makeskycat.conf_$$
echo COL_NAME  = ALPHA_J2000 >> ${TEMPDIR}/makeskycat.conf_$$
echo COL_TTYPE = DOUBLE      >> ${TEMPDIR}/makeskycat.conf_$$
echo COL_HTYPE = FLOAT       >> ${TEMPDIR}/makeskycat.conf_$$
echo COL_COMM = '""'         >> ${TEMPDIR}/makeskycat.conf_$$
echo COL_UNIT = "deg"        >> ${TEMPDIR}/makeskycat.conf_$$
echo COL_DEPTH = 1           >> ${TEMPDIR}/makeskycat.conf_$$
echo COL_NAME  = DELTA_J2000 >> ${TEMPDIR}/makeskycat.conf_$$
echo COL_TTYPE = DOUBLE      >> ${TEMPDIR}/makeskycat.conf_$$
echo COL_HTYPE = FLOAT       >> ${TEMPDIR}/makeskycat.conf_$$
echo COL_COMM = '""'         >> ${TEMPDIR}/makeskycat.conf_$$
echo COL_UNIT = "deg"        >> ${TEMPDIR}/makeskycat.conf_$$
echo COL_DEPTH = 1           >> ${TEMPDIR}/makeskycat.conf_$$
echo COL_NAME  = MAG_AUTO    >> ${TEMPDIR}/makeskycat.conf_$$
echo COL_TTYPE = FLOAT       >> ${TEMPDIR}/makeskycat.conf_$$
echo COL_HTYPE = FLOAT       >> ${TEMPDIR}/makeskycat.conf_$$
echo COL_COMM = '""'         >> ${TEMPDIR}/makeskycat.conf_$$
echo COL_UNIT = '""'         >> ${TEMPDIR}/makeskycat.conf_$$
echo COL_DEPTH = 1           >> ${TEMPDIR}/makeskycat.conf_$$

${P_ASCTOLDAC} -a ${TEMPDIR}/skc_tmp2_$$ \
    -o ${TEMPDIR}/skc_tmp3_$$ \
    -t OBJECTS -c ${TEMPDIR}/makeskycat.conf_$$

${P_LDACTOSKYCAT} -i ${TEMPDIR}/skc_tmp3_$$ -t OBJECTS \
    -k SeqNr ALPHA_J2000 DELTA_J2000 MAG_AUTO \
    -l id_col SeqNr ra_col ALPHA_J2000 dec_col \
    DELTA_J2000 mag_col MAG_AUTO \
    -s > ~/.theli/scripts/theli_mystd.skycat

mkdir -p /$1/$2/cat/ds9cat
mkdir -p /$1/$2/cat/skycat

cp ~/.theli/scripts/theli_mystd.skycat /$1/$2/cat/skycat
cp ~/.theli/scripts/theli_mystd.reg /$1/$2/cat/ds9cat

rm ${TEMPDIR}/skc_tmp[2,3]_$$ ${TEMPDIR}/tmp_scamp_$$ ${TEMPDIR}/tmp2_scamp0_$$
rm ${TEMPDIR}/tmp2_scamp1_$$