#!/bin/bash -uxv

# create catalogues suitable for absolute photometric calibration
# with 'create_abs_photo_info.sh'. The script takes astrometric
# information from a scamp processing and calculates 'correct' sky
# coordinates from objects catalogues (extracted from standard star
# fields). Afterwards the standard star object catalogues are
# merged with a reference standard star sources (e.g. from Landolt,
# Stetson, Sloan ....)

#$1: main directory
#$2: science dir. (the catalogues are assumed to be in $1/$2/cat)
#$3: image extension

INSTDIR=instruments_professional
if [ ! -f ${INSTDIR}/${INSTRUMENT}.ini ]; then
    INSTDIR=instruments_commercial
fi
if [ ! -f ${INSTDIR}/${INSTRUMENT}.ini ]; then
    INSTDIR=~/.theli/instruments_user/
fi
. ${INSTDIR}/${INSTRUMENT:?}.ini


# give meaningful names to command line arguments:
MD=$1
SD=$2
EXTENSION=$3

DIR=`pwd`

if [ "${V_ABSPHOT_STDCAT_INDIRECT}" = "LANDOLT" ]; then
    PHOTCAT=${DIR}"/../photstdcats/LANDOLT.cat"
    MAKESSCCONF="stdphotom_prepare_make_ssc_UBVRI.conf"
fi
if [ "${V_ABSPHOT_STDCAT_INDIRECT}" = "STETSON" ]; then
    PHOTCAT=${DIR}"/../photstdcats/STETSON.cat"
    MAKESSCCONF="stdphotom_prepare_make_ssc_UBVRI.conf"
fi
if [ "${V_ABSPHOT_STDCAT_INDIRECT}" = "STRIPE82_ugriz" ]; then
    PHOTCAT=${DIR}"/../photstdcats/STRIPE82.cat"
    MAKESSCCONF="stdphotom_prepare_make_ssc_ugriz.conf"
fi
if [ "${V_ABSPHOT_STDCAT_INDIRECT}" = "STRIPE82_u'g'r'i'z'" ]; then
    PHOTCAT=${DIR}"/../photstdcats/STRIPE82_primed.cat"
    MAKESSCCONF="stdphotom_prepare_make_ssc_ugriz.conf"
fi
if [ "${V_ABSPHOT_STDCAT_INDIRECT}" = "MKO_JHK" ]; then
    PHOTCAT=${DIR}"/../photstdcats/MKO.cat"
    MAKESSCCONF="stdphotom_prepare_make_ssc_JHK.conf"
fi
if [ "${V_ABSPHOT_STDCAT_INDIRECT}" = "MKO_LM" ]; then
    PHOTCAT=${DIR}"/../photstdcats/MKO_LM.cat"
    MAKESSCCONF="stdphotom_prepare_make_ssc_LM.conf"
fi
if [ "${V_ABSPHOT_STDCAT_INDIRECT}" = "HUNT_JHK" ]; then
    PHOTCAT=${DIR}"/../photstdcats/HUNT.cat"
    MAKESSCCONF="stdphotom_prepare_make_ssc_JHK.conf"
fi
if [ "${V_ABSPHOT_STDCAT_INDIRECT}" = "2MASSfaint_JHK" ]; then
    PHOTCAT=${DIR}"/../photstdcats/2MASSfaint.cat"
    MAKESSCCONF="stdphotom_prepare_make_ssc_JHK.conf"
fi
if [ "${V_ABSPHOT_STDCAT_INDIRECT}" = "PERSSON_JHKKs" ]; then
    PHOTCAT=${DIR}"/../photstdcats/PERSSON.cat"
    MAKESSCCONF="stdphotom_prepare_make_ssc_JHKKs.conf"
fi
if [ "${V_ABSPHOT_STDCAT_INDIRECT}" = "JAC_YJHKLM" ]; then
    PHOTCAT=${DIR}"/../photstdcats/JAC.cat"
    MAKESSCCONF="stdphotom_prepare_make_ssc_YJHKLM.conf"
fi
if [ "${V_ABSPHOT_STDCAT_INDIRECT}" = "PERSSON_RED_JHKKs" ]; then
    PHOTCAT=${DIR}"/../photstdcats/PERSSON_RED.cat"
    MAKESSCCONF="stdphotom_prepare_make_ssc_JHKKs.conf"
fi

nobj_total=0

for CHIP in ${!#}
do
  CATS=`find ${MD}/${SD}/cat -name \*_${CHIP}${EXTENSION}.cat`

  for CAT in ${CATS}
  do
    BASE=`basename ${CAT} .cat`
    HEADBASE=`basename ${CAT} ${EXTENSION}.cat`

    # get 'correct' astrometric scamp information to the catalogues:    
    ${S_APLASTROMSCAMP} -i ${MD}/${SD}/cat/${BASE}.cat \
         -o ${MD}/${SD}/cat/${BASE}_corr.cat \
         -s ${MD}/${SD}/headers/${HEADBASE}.head \
         -t OBJECTS -p Xpos Ypos -r Ra Dec

    ${P_MAKEJOIN} -i /${MD}/${SD}/cat/${BASE}_corr.cat \
                  -o ${TEMPDIR}/tmp.cat0_$$ \
		  -c ${DATACONF}/stdphotom_prepare_make_join.conf
    # calculate object magnitudes with a magzeropoint of 0 and
    # an exposure time normalisation of 1s (1.08574 is 2.5 / log(10)):
    ${P_LDACCALC} -i ${TEMPDIR}/tmp.cat0_$$ \
                  -o ${TEMPDIR}/tmp.cat00_$$ \
                  -t OBJECTS -c "(-1.08574*log(FLUX_AUTO/EXPTIME));" \
                  -n MAG_AUTO_corr "exposure time corr. MAG_AUTO" -k FLOAT
    ${P_LDACRENTAB} -i ${TEMPDIR}/tmp.cat00_$$ \
                    -o ${TEMPDIR}/tmp.cat1_$$ -t OBJECTS STDTAB
    ${P_LDACRENKEY} -i ${TEMPDIR}/tmp.cat1_$$ -o ${TEMPDIR}/tmp.cat2_$$ \
                    -t STDTAB -k A_WORLD A_WCS B_WORLD B_WCS \
                                 THETA_J2000 THETAWCS
    ${P_ASSOCIATE} -i ${TEMPDIR}/tmp.cat2_$$ ${PHOTCAT} \
		   -o ${TEMPDIR}/obj_tmp.cat3_$$ ${TEMPDIR}/ref_tmp.cat4_$$ \
                   -t STDTAB -c ${DATACONF}/stdphotom_prepare_associate.conf

    # if no objects found, continue with next catalog
    nobj=`${P_LDACTOASC} -i ${TEMPDIR}/obj_tmp.cat3_$$ -t STDTAB -k Pair_1 -b | \
	${P_GAWK} 'BEGIN {s=0} {if ($1>0) s++} END {print s}'`
    if [ "${nobj}" = "0" ]; then
	continue
    fi

    ${P_LDACFILTER} -i ${TEMPDIR}/obj_tmp.cat3_$$ -o ${TEMPDIR}/obj_tmp.cat5_$$ \
                    -c "(Pair_1>0);" -t STDTAB

    if [ "$?" -eq "0" ]; then
        # if no objects found, continue with next catalog
	nobj=`${P_LDACTOASC} -i ${TEMPDIR}/ref_tmp.cat4_$$ -t STDTAB -k Pair_0 -b | \
	    ${P_GAWK} 'BEGIN {s=0} {if ($1>0) s++} END {print s}'`
	if [ "${nobj}" = "0" ]; then
	    continue
	fi
	${P_LDACFILTER} -i ${TEMPDIR}/ref_tmp.cat4_$$ \
            -o ${TEMPDIR}/ref_tmp.cat6_$$ -c "(Pair_0>0);" -t STDTAB
	${P_ASSOCIATE} -i ${TEMPDIR}/obj_tmp.cat5_$$ ${TEMPDIR}/ref_tmp.cat6_$$ \
            -o ${TEMPDIR}/obj_tmp.cat7_$$ ${TEMPDIR}/ref_tmp.cat8_$$ \
      	    -t STDTAB -c ${DATACONF}/stdphotom_prepare_associate.conf
	${P_MAKESSC} -i ${TEMPDIR}/obj_tmp.cat7_$$ ${TEMPDIR}/ref_tmp.cat8_$$ \
            -o /${MD}/${SD}/cat/${BASE}_merg.cat \
            -t STDTAB \
            -c ${DATACONF}/${MAKESSCCONF}

	nobj_total=$((${nobj_total} + ${nobj}))
    fi
  done
  ${P_LDACPASTE} -i /${MD}/${SD}/cat/*_${CHIP}${EXTENSION}_merg.cat \
                 -o /${MD}/${SD}/cat/chip_${CHIP}_merg.cat -t PSSC
done

# if no overlapping sources were found:

if [ "${nobj_total}" = "0" ]; then
    echo `${P_ERRTEST} 0` Your standard star field does not overlap with ${V_ABSPHOT_STDCAT_INDIRECT}, or the standard stars could not be identified/matched.
    exit 1;
fi

# clean up
rm  ${TEMPDIR}/*_$$
