#!/bin/bash -xv

#$1: main directory (filter)
#$2: Science directory
#$3: chips to be processed

# preliminary work:

SCRIPTDIR=`pwd`

INSTDIR=instruments_professional
if [ ! -f ${INSTDIR}/${INSTRUMENT}.ini ]; then
    INSTDIR=instruments_commercial
fi
if [ ! -f ${INSTDIR}/${INSTRUMENT}.ini ]; then
    INSTDIR=~/.theli/instruments_user/
fi
. ${INSTDIR}/${INSTRUMENT:?}.ini

. bash.include

# the resultdir is where the output coadded images
# will go. If ONE image of the corresponding chip
# is a link the image will go to THIS directory
for CHIP in $3
do
  RESULTDIR[${CHIP}]="$1/$2"
done

# we create object-subtracted images to avoid dark regions
# around bright objects in later images

if [ "${V_CAL_WINDOWMETHOD}" = "0" ]; then
    WINDOWMETHOD="superflat"
fi
if [ "${V_CAL_WINDOWMETHOD}" = "1" ]; then
    WINDOWMETHOD="defringe"
fi
if [ "${V_CAL_WINDOWMETHOD}" = "2" ]; then
    WINDOWMETHOD="subtract"
fi

SUB=""

##########################################################
#        create object-subtracted images first           #
##########################################################

if [ "${V_CAL_DETECTMINAREA}" != "" ] && [ "${V_CAL_DETECTTHRESH}" != "" ]; then
  SUB=".mask"
  for CHIP in $3
  do
    ls -1 /${RESULTDIR[${CHIP}]}/*_${CHIP}OFC.fits \
	  > ${TEMPDIR}/images-objects_$$

    NUMTHREAD=`${SCRIPTDIR}/get_nthreads.sh ${NCHIPS} ${NPARA} ${CHIP}`

    cat ${TEMPDIR}/images-objects_$$ |\
    {
      while read file
	do
	  BASE=`basename ${file} .fits`

	  # check if a mask image already exists, and if it was created from a collapse correction
	  from_collapsed="no"
	  if [ -e /$1/$2/MASK_IMAGES/${BASE}.mask.fits ]; then
	      from_collapsed=`get_key /$1/$2/MASK_IMAGES/${BASE}.mask.fits COLLAPSE`
	  fi

	  # check if a mask image already exists, and if it was created from a collapse correction,
	  # then just copy it; else re-create from scratch
	  if [ -f /$1/$2/MASK_IMAGES/${BASE}.mask.fits ] && [ "${from_collapsed}" = "yes" ]; then
	      # transfer the masks in MASK_IMAGES
	      ln -s /$1/$2/MASK_IMAGES/${BASE}.mask.fits $1/$2/${BASE}"${SUB}.fits"
	  else

	      # if large parts of the image are empty or vignetted, then we can use ds9-type
	      # polygon masks to mask these pixels. In addition, we need to force sextractor
	      # to assume a fixed sky background, as otherwise the large intensity variations
	      # between good image areas and vignetted areas screw up the sextractor background 
	      # modelling
	      region=""
	      backstring=""
	      if [ -f ${INSTDIR}/${INSTRUMENT}.reg ]; then
		  region="-p ${INSTDIR}/${INSTRUMENT}.reg"
		  # estimate the background, ignoring blanked pixels
		  background=`${P_FITSSTAT} -i ${file} -m median -q`
		  backstring="-BACK_TYPE MANUAL -BACK_VALUE ${background}"
	      fi

	      ${P_SEX} ${file} -c ${DATACONF}/image-objects.sex\
		  -CHECKIMAGE_NAME ${RESULTDIR[${CHIP}]}/${BASE}"${SUB}.fits" \
		  -DETECT_MINAREA ${V_CAL_DETECTMINAREA} \
		  -DETECT_THRESH ${V_CAL_DETECTTHRESH} \
		  -ANALYSIS_THRESH ${V_CAL_DETECTTHRESH} \
		  -BACK_SIZE 256 \
		  -FILTER ${V_CAL_SEXFILTER} \
		  -NTHREADS ${NUMTHREAD} \
		  ${backstring}

	      # set pixels in the input image to a very large negative value
	      # if they are (a) part of an object, or (b) masked by a ds9 region file;
	      # we can safely overwrite the comparison image (-t) with the result (-o)
	      ${P_FITSPOLYGON} \
		  -i ${file} \
		  -t ${RESULTDIR[${CHIP}]}/${BASE}"${SUB}.fits" \
		  -o ${RESULTDIR[${CHIP}]}/${BASE}"${SUB}.fits" \
		  -u 0. \
		  -v -1.0e9 \
		  ${region}
	  fi

	  if [ "${RESULTDIR[${CHIP}]}" != "$1/$2" ]; then
	      ln -s ${RESULTDIR[${CHIP}]}/${BASE}"${SUB}.fits" \
		  $1/$2/${BASE}"${SUB}.fits"
	  fi
	done
    }
  done
fi


##########################################################
#              now calculate the superflat               #
##########################################################

# if the superflat is calculated from the entire set of images
if [ "${V_CAL_WINDOWSIZE}" = "" ]; then
    V_CAL_WINDOWSIZE=0
fi
if [ "${V_CAL_WINDOWSIZE}" = "0" ]; then
    for CHIP in $3
	do
	if [ "${RESULTDIR[${CHIP}]}" != "$1/$2" ]; then
	    ln -s ${RESULTDIR[${CHIP}]}/$2_${CHIP}.fits $1/$2/$2_${CHIP}.fits
	fi
	
        # do statistics from all science frames.
        # This is necessary to get the mode
        # of the combination from all images, so that
        # the combination where images have been excluded
        # can be scaled accordingly.
	${P_IMSTATS} -d 6 `ls /$1/$2/*_${CHIP}OFC${SUB}.fits` \
	    -o ${TEMPDIR}/science_images_$$ -t -1e8 1e8 
	
	RESULTMODE=`${P_GAWK} 'BEGIN {mean=0.0; n=0} ($1!="#") {
         n=n+1; mean=mean+$2} END {print mean/n}' ${TEMPDIR}/science_images_$$`
	
        # modify the input list of images
        # in case we have to reject files for the superflat:
	if [ -s /$1/$2/superflat_exclusion ]; then
	    ${P_GAWK} 'BEGIN {nex = 0; while (getline <"/'$1/$2'/superflat_exclusion" > 0) {
             ex[nex]=$1; nex++; }}
             {exclude = 0; i=0;
              while (i<nex) {
                if (index($1, ex[i])!=0) {
                  exclude = 1;
                }
	          i++;
              }
              if(exclude == 0) {print $0}}' ${TEMPDIR}/science_images_$$ > \
		  ${TEMPDIR}/science_coadd_images_$$
	else
	    cp ${TEMPDIR}/science_images_$$ ${TEMPDIR}/science_coadd_images_$$
	fi
	
	# exit if only one image contributes to the stack
	stacksize=`${P_GAWK} 'BEGIN{count=0} {if ($1 != "#") count=count+1} END {print count-'"${V_CAL_SCIENCENLOW}"'-'"${V_CAL_SCIENCENHIGH}"'}' \
	    ${TEMPDIR}/science_coadd_images_$$`
	if [ ${stacksize} -lt 2 ]; then
	    ${P_ERRTEST} 0
	    echo "Only 1 or less images contribute effectively to the superflat!"
	    echo "Consider reducing the nlow/nhigh rejection parameter."
	    exit
	fi
	
        # do the combination; do not include images
        # listed in the /$1/$2/superflat_exclusion file
	${P_IMCOMBFLAT_IMCAT} -i ${TEMPDIR}/science_coadd_images_$$ \
	    -o ${RESULTDIR[${CHIP}]}/$2_${CHIP}.fits \
	    -c ${V_CAL_COMBINEMETHOD} -s 1 -e ${V_CAL_SCIENCENLOW} \
	    ${V_CAL_SCIENCENHIGH} -m ${RESULTMODE} -t -1e8 1e8
	
	if [ "${SUB}" = ".mask" ]; then
	    if [ ! -d /$1/$2/MASK_IMAGES ]; then
		mkdir /$1/$2/MASK_IMAGES
	    fi

	    # clean up
	    ls /$1/$2/*_${CHIP}OFC.mask.fits > ${TEMPDIR}/science_coadd_images2_$$
	    exec < ${TEMPDIR}/science_coadd_images2_$$
	    while read file
	    do
		BASE=`basename ${file}`
		# check if the mask is a link, in that case delete it, else mv
		test /$1/$2/${BASE} -ef /$1/$2/MASK_IMAGES/${BASE} && \
		    rm /$1/$2/${BASE} || mv /$1/$2/${BASE} /$1/$2/MASK_IMAGES/${BASE}
	    done
	fi
    done
fi

# if the superflat is calculated from a small subset ("window") 
if [ "${V_CAL_WINDOWSIZE}" != "0" ]; then

    if [ ! -d /$1/$2/BACKGROUND ]; then
	mkdir /$1/$2/BACKGROUND
    fi

    for CHIP in $3
    do
	if [ "${RESULTDIR[${CHIP}]}" != "$1/$2" ]; then
	    ln -s ${RESULTDIR[${CHIP}]}/$2_${CHIP}.fits $1/$2/$2_${CHIP}.fits
	fi

	NUMTHREAD=`${SCRIPTDIR}/get_nthreads.sh ${NCHIPS} ${NPARA} ${CHIP}`

	k=1
	ls -1 /$1/$2/*_${CHIP}OFC.fits > ${TEMPDIR}/images_to_flat_$$
	
	get_statminsection

	# loop over all images
	cat ${TEMPDIR}/images_to_flat_$$ |\
	{
	  while read individualfile
	  do

	    # get the image names from which to calculate the superflat
	    num=`${P_FIND} /$1/$2/ -maxdepth 1 -name \*_${CHIP}OFC${SUB}.fits | wc | ${P_GAWK} '{print $1}'`
	    win=${V_CAL_WINDOWSIZE}
	    ind=${k}
	    # include or exclude the centre image of the window
	    if [ "${V_CAL_EXCLUDECENTERIMAGE}" = "TRUE" ]; then
		images=`${P_FIND} /$1/$2/ -maxdepth 1 -name \*_${CHIP}OFC${SUB}.fits | ${P_SORT} | ${P_GAWK} 'BEGIN {
                l='"${win}"'
                i='"${ind}"'
                n='"${num}"'
                j=1
                start=i-l
                test=0
                if (i-l<1) {
                   start=1
                   test=i-l-1
                }
                if (i+l>n) {
                   start=n-2*l
                   test=i+l-n
                }
	      } {
              if (NR>=start && j<=2*l+1) {
                 if (j != l+1+test) print
                 j++
              } }'`
	    else
		images=`${P_FIND} /$1/$2/ -maxdepth 1 -name \*_${CHIP}OFC${SUB}.fits | ${P_SORT} | ${P_GAWK} 'BEGIN {
                l='"${win}"'
                i='"${ind}"'
                n='"${num}"'
                j=1
                start=i-l
                if (i-l<1) start=1
                if (i+l>n) start=n-2*l } {
              if (NR>=start && j<=2*l+1) {
                  print
                  j++
              } }'`
	    fi
	    
            # do statistics from all science frames.
            # This is necessary to get the mode
            # of the combination from all images, so that
            # the combination where images have been excluded
            # can be scaled accordingly.
	    ${P_IMSTATS} -d 6 ${images} \
		-o ${TEMPDIR}/science_images_$$ -t -1e8 1e8

	    RESULTMODE=`${P_GAWK} 'BEGIN {mean=0.0; n=0} ($1!="#") {
             n=n+1; mean=mean+$2} END {print mean/n}' ${TEMPDIR}/science_images_$$`
	
            # modify the input list of images
            # in case we have to reject files for the superflat:
	    if [ -s /$1/$2/superflat_exclusion ]; then
		${P_GAWK} 'BEGIN {nex = 0; while (getline <"/'$1/$2'/superflat_exclusion" > 0) {
                 ex[nex]=$1; nex++; }}
                 {exclude = 0; i=0;
                  while (i<nex) {
                    if (index($1, ex[i])!=0) {
                      exclude = 1;
                    }
	              i++;
                  }
                  if(exclude == 0) {print $0}}' ${TEMPDIR}/science_images_$$ >\
			${TEMPDIR}/science_coadd_images_$$
	    else
		cp ${TEMPDIR}/science_images_$$ ${TEMPDIR}/science_coadd_images_$$
	    fi
	
            # exit if only one image contributes to the stack
	    stacksize=`${P_GAWK} 'BEGIN{count=0} {if ($1 != "#") count=count+1} END {print count-'"${V_CAL_SCIENCENLOW}"'-'"${V_CAL_SCIENCENHIGH}"'}' \
		${TEMPDIR}/science_coadd_images_$$`
	    if [ ${stacksize} -lt 2 ]; then
		${P_ERRTEST} 0
		echo "Only 1 or less images contribute effectively to the superflat!"
		echo "Consider reducing the nlow/nhigh rejection parameter."
		exit
	    fi

            # do the combination; do not include images
            # listed in the /$1/$2/superflat_exclusion file
	    ${P_IMCOMBFLAT_IMCAT} -i ${TEMPDIR}/science_coadd_images_$$\
		-o ${RESULTDIR[${CHIP}]}/$2_${CHIP}.fits -c ${V_CAL_COMBINEMETHOD} \
		-s 1 -e ${V_CAL_SCIENCENLOW} ${V_CAL_SCIENCENHIGH} -m ${RESULTMODE}  -t -1e8 1e8
	    
	    # now apply the superflat (superflat, defringe, or subtract)	    
	    if [ "${WINDOWMETHOD}" != "subtract" ]; then
		FILE=`ls /$1/$2/$2_${CHIP}.fits`
		if [ -L ${FILE} ]; then
		    LINK=`${P_READLINK} ${FILE}`
		    BASE=`basename ${LINK} .fits`
		    DIR=`dirname ${LINK}`
		    ln -s ${DIR}/$2_${CHIP}_illum.fits $1/$2/$2_${CHIP}_illum.fits
		    ln -s ${DIR}/$2_${CHIP}_fringe.fits $1/$2/$2_${CHIP}_fringe.fits
		fi
		
		${P_SEX} /$1/$2/$2_${CHIP}.fits -c ${CONF}/illumfringe_back.sex \
		    -CHECKIMAGE_NAME ${RESULTDIR[${CHIP}]}/$2_${CHIP}_illum.fits \
		    -BACK_SIZE ${V_ILLUMFRINGEBACKSIZE} -NTHREADS ${NUMTHREAD}
		${P_SEX} /$1/$2/$2_${CHIP}.fits -c ${CONF}/illumfringe_fringe.sex \
		    -CHECKIMAGE_NAME ${RESULTDIR[${CHIP}]}/$2_${CHIP}_fringe.fits \
		    -BACK_SIZE ${V_ILLUMFRINGEBACKSIZE} -NTHREADS ${NUMTHREAD}

		# only superflat the data
		if [ "${WINDOWMETHOD}" = "superflat" ]; then
		    if [ "${V_SF_SMOOTH}" = "SMOOTHED" ]; then
			FLATIMAGE=/$1/$2/$2_${CHIP}_illum.fits
		    fi
		    if [ "${V_SF_SMOOTH}" = "UNSMOOTHED" ]; then
			FLATIMAGE=/$1/$2/$2_${CHIP}.fits
		    fi
		    
		    ${P_IMRED_ECL:?} ${individualfile} \
			-c ${CONF}/imred.conf \
			-MAXIMAGES ${NFRAMES} \
			-OVERSCAN N \
			-BIAS N \
			-FLAT Y \
			-FLAT_IMAGE ${FLATIMAGE} \
			-FLAT_THRESHHOLD 0.1 \
			-FLAT_SCALE N \
			-COMBINE N \
			-OUTPUT Y \
			-OUTPUT_DIR /$1/$2/ \
			-OUTPUT_SUFFIX S.fits \
			-TRIM N \
			-STATSSEC ${statminsection}
		fi

		# superflat and defringe the data
		if [ "${WINDOWMETHOD}" = "defringe" ]; then
		    FLATIMAGE=/$1/$2/$2_${CHIP}_illum.fits
		    FRINGEIMAGE=/$1/$2/$2_${CHIP}_fringe.fits
		    ${P_IMRED_ECL:?} ${individualfile} \
			-c ${CONF}/imred.conf \
			-MAXIMAGES ${NFRAMES} \
			-OVERSCAN N \
			-BIAS N \
			-FLAT Y \
			-FLAT_IMAGE ${FLATIMAGE} \
			-FLAT_THRESHHOLD 0.1 \
			-FLAT_SCALE N \
			-FRINGE Y \
			-FRINGE_IMAGE /$1/$2/$2_${CHIP}_fringe.fits \
			-FRINGE_REFIMAGE /$1/$2/$2_${CHIP}_illum.fits \
			-COMBINE N \
			-OUTPUT Y \
			-OUTPUT_DIR /$1/$2/ \
			-OUTPUT_SUFFIX SF.fits \
			-TRIM N \
			-STATSSEC ${statminsection}
		fi
	    fi

	    # subtract superflat
	    if [ "${WINDOWMETHOD}" = "subtract" ]; then
		FRINGEIMAGE=/$1/$2/$2_${CHIP}.fits
		${P_IMRED_ECL:?} ${individualfile} \
		    -c ${CONF}/imred.conf \
		    -MAXIMAGES ${NFRAMES} \
		    -OVERSCAN N \
		    -BIAS N \
		    -FLAT N \
		    -FRINGE Y \
		    -FRINGE_IMAGE /$1/$2/$2_${CHIP}.fits \
		    -FRINGE_REFIMAGE /$1/$2/$2_${CHIP}.fits \
		    -COMBINE N \
		    -OUTPUT Y \
		    -OUTPUT_DIR /$1/$2/ \
		    -OUTPUT_SUFFIX U.fits \
		    -TRIM N \
		    -STATSSEC ${statminsection}

		# write the sky background into the header
		skyexist=`get_key ${individualfile} SKYVALUE`
		if [ "${skyexist}"_A = "_A" ]; then
		    NEXTDUMMY=`${P_DFITS} ${individualfile} | grep DUMMY | \
			${P_GAWK} 'BEGIN {nrmin=1000+0} {
                        if (index($1,"DUMMY")==1) {
			nr=substr($1,6)+0; if(nr<nrmin) {nrmin=nr}}}
                        END {if(NR>0) {print nrmin}}'`
		    sky=`${P_IMSTATS} -d 6 ${FILE} -t -1e8 1e8 | \
			${P_GAWK} '($1 != "#") {print $2}'`
		#    sky=`${P_IMSTATS} -d 6 /$1/$2/$2_${CHIP}.fits -t -1e8 1e8 | \
		#	${P_GAWK} '($1 != "#") {print $2}'`
		    BASE=`basename ${individualfile} U.fits`
		    ${P_REPLACEKEY} ${individualfile} "SKYVALUE= ${sky}" DUMMY${NEXTDUMMY}
		    ${P_REPLACEKEY} /$1/$2/${BASE}U.fits "SKYVALUE= ${sky}" DUMMY${NEXTDUMMY}
		fi

		BASE=`basename ${individualfile} U.fits`
		if [ "${RESULTDIR[${CHIP}]}" != "$1/$2" ]; then
		    ln -s ${RESULTDIR[${CHIP}]}/${BASE}U.fits \
			/$1/$2/${BASE}U.fits
		fi
	    fi

	    BASETMP=`basename ${individualfile} .fits`
	    mv /$1/$2/$2_${CHIP}.fits $1/$2/BACKGROUND/${BASETMP}_sky.fits
	    test -e /$1/$2/$2_${CHIP}_illum.fits && mv /$1/$2/$2_${CHIP}_illum.fits $1/$2/BACKGROUND/${BASETMP}_skyillum.fits
	    test -e /$1/$2/$2_${CHIP}_fringe.fits && mv /$1/$2/$2_${CHIP}_fringe.fits $1/$2/BACKGROUND/${BASETMP}_skyfringe.fits

	    k=$((${k}+1))
	  done
	}
	
	if [ "${SUB}" = ".mask" ]; then
	    if [ ! -d /$1/$2/MASK_IMAGES ]; then
		mkdir /$1/$2/MASK_IMAGES
	    fi
	    # clean up
	    ls /$1/$2/*_${CHIP}OFC.mask.fits > ${TEMPDIR}/science_coadd_images2_$$
	    exec < ${TEMPDIR}/science_coadd_images2_$$
	    while read file
	    do
		BASE=`basename ${file}`
		# check if the mask is a link, in that case delete it, else mv
		test /$1/$2/${BASE} -ef /$1/$2/MASK_IMAGES/${BASE} && \
		    rm /$1/$2/${BASE} || mv /$1/$2/${BASE} /$1/$2/MASK_IMAGES/${BASE}
	    done
	    mv /$1/$2/*_${CHIP}OFC${SUB}.fits /$1/$2/MASK_IMAGES/
	fi
	if [ ! -d /$1/$2/OFC_IMAGES ]; then
	    mkdir /$1/$2/OFC_IMAGES
	fi
	mv /$1/$2/*_${CHIP}OFC.fits /$1/$2/OFC_IMAGES
    done
fi
