/*---------------------------------------------------------------------------

   File name   : preprocess.c
   Author      : T. Erben
   Created on  : August 2003
   Description : Image preprocessing

*--------------------------------------------------------------------------*/

/*
        $Id: preprocess.c,v 1.22 2012-01-31 13:51:30 thomas Exp $
        $Author: thomas $
        $Date: 2012-01-31 13:51:30 $
        $Revision: 1.22 $
 */

/*---------------------------------------------------------------------------
                                                     History comments
   ---------------------------------------------------------------------------*/

/*
   14.09.03:
   included set_verbose(1) to get comment messages printed by
   e_comment etc. The messages previously appeared on Linux SuSE
   but not on Sun and Linux Debian

   22.10.03:
   included the possibility to give a "threshhold" for
   the flatfield. Pixels that have an absolute value below
   this threshhold get a value of 0 in the preprocessed
   frame. Hence, pixels with a very high modulus in the
   preprocessed frame are avoided.

   16.01.05:
   I substituted the median estimation by 'image_getmedian_vig'
   by 'image_getmedian_vig_lowhighreject' at several places.
   As low and high rejection thresholds, -60000.0 and 60000.0
   are hardcoded in the moment (has to be made configurable).
   The same is true for the default return value which is
   -60000.0 right now.

   23.01.2005:
   I changed the location of the actual prefstruct definition
   to avoid double existence of it during the link process.

   27.01.2005:
   I added global version number information

   28.01.2005:
   I slightly changed the output format for the global
   version number

   30.01.2005:
   I largely rewrote the code:
   A new parameter MAXIMAGES can be given on the command line.
   It gives the maximum number of images to be loaded; if more
   images than this limit is given, images are loaded in chunks
   of MAXIMAGES and processed. We call the processing of one
   chunk a 'run'. In this way we avoid memory problems
   with large amounts of images. An image coaddition is NO LONGER
   possible if more than one run is to be processed.

   16.02.2005:
   - I corrected a bug in the saving of output images. If several
     runs were processed, images from all runs except the first one
     got wrong image headers.
   - I removed the superfluous definition of 'give_date'
   - frames are now loaded with image_loadext instead with
     image_load. The latter memorymapped the images what led
     to problems if the fringescale image was the same than
     the flatfield that is usually being rescaled.

   08.04.2005:
   The number of input images is now dynamically allocated and
   no longer hardwired. The same for all variables depending on
   the number of input images.

   09.04.2005:
   Warning messages are given if the coadded image consists of
   the default value, i.e. if the number of input images minus
   the rejected ones is too small for a meaningful median estimation.

   16.06.2006:
   In the construction of the output name of images, the output directory
   now always ends with a '/'.

   07.03.2007:
   I added OUTPUT_BITPIX and COMBINE_BITPIX keys that allow
   configuration of the precision of preprocessed and stacked images.

   19.10.2007:
   correction of a typo in a history message.

   20.11.2009:
   Printing of the usage message is now a function - the handling in
   a 'define string' was too cumbersome for the very long usage
   message.

   09.05.2010:
   I included the option to turn off frimge rescaling (request
   from Mischa)

   28.09.2010:
   I corrected a major bug in the fringe frame subtraction. If no
   fringe rescaling was requested fringes were not corrected at all.

   31.01.2012:
   I corrected a memory allocation error for array 'scalemedian'.
 */

/*---------------------------------------------------------------------------
                                                                Includes
   ---------------------------------------------------------------------------*/


#include <stdio.h>
#include <string.h>
#include <math.h>

#include "global_vers.h"
#include "eclipse.h"
#include "wfip_lib.h"
#include "prefs_imred.h"

/*---------------------------------------------------------------------------
                                                                Defines
   ---------------------------------------------------------------------------*/

#define MIN(x, y)    ((x) < (y) ? (x) : (y))

static void printUsage(void);

prefstruct prefs;

/*---------------------------------------------------------------------------
                                                                        Main
   ---------------------------------------------------------------------------*/

int main(int argc, char *argv[])
{
  int a, narg, nim, i, j, run, part = 0, subpart;
  int nruns;               /* the number of runs to be performed:
                              this is the number of input images
                              divided by maximages */
  int nimarun;             /* the number of images processed
                              in the actual run */
  char **argkey, **argval, *str;
  size_t maxargstringlength = 256;

  cube_t *      i_cube    = NULL;
  image_t *     corrected = NULL;
  image_t *     stacked   = NULL;
  image_t *     bias      = NULL;
  image_t *     flat      = NULL;
  image_t *     flatscale = NULL;
  image_t *     fringe    = NULL;
  image_t *     fringeref = NULL;
  char          name_o[MAXCHAR];
  char          tmpstr[MAXCHAR];
  qfits_header *fh;
  qfits_header *fh_stack;
  history *     hs;
  pixelvalue    maxflatscale = 0.0;
  pixelvalue *  combinemedian;
  pixelvalue *  scalemedian;
  pixelvalue *  fringescaling;
  pixelvalue    meanmedian, combinescaling;
  pixelvalue    fringerefmedian = 0, mediantmp;

  /* pixels below lo_rej and above hi_rej
     are not taken into account
     when estimating medians */
  pixelvalue lo_rej = -60000.0;
  pixelvalue hi_rej = 60000.0;

  narg          = nim = 0;
  prefs.ninname = 0;

  /* the number of config parameters and input image
     names cannot be larger than the number of command
     line arguments
   */
  QMALLOC(argkey, char *, argc);
  QMALLOC(argval, char *, argc);
  QMALLOC(prefs.inname, char *, argc);
  strcpy(prefs.prefs_name, "default.imred");

  /* read command line arguments */

  if (argc < 2) {
    printUsage();
    exit(1);
  }

  for (a = 1; a < argc; a++) {
    if (argv[a][0] == '-') {
      if (strlen(argv[a]) < 3) {
        switch ((int)tolower((int)argv[a][1])) {
          /*-------- Config filename */
        case 'h':    
          printUsage();
          exit(0); 
          break;
        case 'd':    
          dumpprefs();
          exit(0);
          break;
        case 'c':    
            if (a < (argc - 1)) {
              strncpy(prefs.prefs_name, argv[++a], MAXCHAR - 1);
            } else {
              printUsage();
              exit(1);
            }
          break;
        default:    
          printUsage();
          exit(1);
          break;
        }
      } else {
        /*------ Config parameters */
        argkey[narg] = &argv[a][1];
        argval[narg] = argv[++a];

        /* determine the maximum stringlength for command line
           arguments; the factor 2 represents a conservative
           upper limit. The command line arguments consist of
           key AND argument */
        if (2 * strlen(argval[narg]) > maxargstringlength) {
          maxargstringlength = 2 * strlen(argval[narg]);
        }
        narg++;
      }
    } else {
      /*---- The input image filename(s) */
      for (; (a < argc) && (*argv[a] != '-'); a++) {
        for (str = NULL; (str = strtok(str ? NULL : argv[a], notokstr)); nim++) {
          prefs.inname[nim] = str;
        }
      }
      prefs.ninname = nim;
      a--;
    }
  }

  readprefs(prefs.prefs_name, argkey, argval, narg, maxargstringlength);

  /* allocate memory for variables depending on the
     number of images */
  QMALLOC(combinemedian, pixelvalue, prefs.maximages);
  QMALLOC(scalemedian, pixelvalue, prefs.nflatscaleimage);
  QMALLOC(fringescaling, pixelvalue, prefs.maximages);

  free(argkey);
  free(argval);

  /*
   * Initialize eclipse environment.
   */
  eclipse_init();
  set_verbose(1);   /* print comment messages */

  e_comment(0, "-> Part %d: Preliminary work:", ++part);
  subpart = 1;

  /* determine number of necessary runs */
  nruns = (prefs.ninname - 1) / prefs.maximages;

  if (nruns > 0 && prefs.combineflag == 1) {
    e_warning("cannot perform stacking with several runs:");
    e_warning("no coaddition will be done!!");
    prefs.combineflag = 0;
  }

  /* sanity check whether there is something to do at all */
  if (prefs.ninname == 0 || (prefs.overscanflag == 0 && prefs.trimflag == 0 &&
                             prefs.biasflag == 0 && prefs.flatflag == 0 &&
                             prefs.fringeflag == 0 && prefs.combineflag == 0)) {
    e_warning("nothing do do!! exiting!!");
    return(0);
  }

  /* sanity checks on BITPIX flags if necessary */
  if (prefs.outputpreprocflag == 1 &&
      (prefs.outputbitpix != -32 && prefs.outputbitpix != 16)) {
    e_error("Only BITPIX=16 or BITPIX=-32 supported for output images.");
    return(-1);
  }

  if (prefs.combineflag == 1 &&
      (prefs.combinebitpix != -32 && prefs.combinebitpix != 16)) {
    e_error("Only BITPIX=16 or BITPIX=-32 supported for stacked images.");
    return(-1);
  }

  /* load necessary calibration frames:
     There is no need to load them for
     every run
   */
  if (prefs.biasflag == 1) {
    e_comment(1, "-> %d.%d: Load Bias Frame %s", part, 
              subpart++, prefs.biasimage);

    /* load bias frame */
    bias = image_loadext(prefs.biasimage, 0, 0, 0);

    if (bias == NULL) {
      e_error("cannot load bias [%s]: aborting", prefs.biasimage);
      return(-1);
    }
  }

  if (prefs.flatflag == 1) {
    e_comment(1, "-> %d.%d: Load Flatfield Frame %s", part,
              subpart++, prefs.flatimage);

    /* load flatfield */
    flat = image_loadext(prefs.flatimage, 0, 0, 0);

    if (flat == NULL) {
      e_error("cannot load flat [%s]: aborting", prefs.flatimage);
      return(-1);
    }

    /* determine flatfield normalisation */
    maxflatscale = image_getmedian_vig_lowhighreject(flat,
                                                     prefs.statssec[0],
                                                     prefs.statssec[2],
                                                     prefs.statssec[1],
                                                     prefs.statssec[3],
                                                     lo_rej,
                                                     hi_rej,
                                                     lo_rej);

    if (maxflatscale < prefs.flatthreshhold) {
      e_error("Flat image [%s] has median %f which is below the threshold %f: aborting",
              prefs.flatimage, maxflatscale, prefs.flatthreshhold);
      return(-1);
    }


    e_comment(2, "Flat %s has median %f", prefs.flatimage,
              maxflatscale);

    if (prefs.flatscaleflag == 1) {
      /* determine scale factor:
         it is the maximum gain from all the different
         chips
       */
      maxflatscale = 0.0;
      for (i = 0; i < prefs.nflatscaleimage; i++) {
        flatscale = image_loadext(prefs.flatscaleimage[i], 0, 0, 0);

        if (flatscale == NULL) {
          e_error("cannot load image [%s]: aborting",
                  prefs.flatscaleimage[i]);
          return(-1);
        }
        scalemedian[i] = image_getmedian_vig_lowhighreject(flatscale,
                                                           prefs.statssec[0],
                                                           prefs.statssec[2],
                                                           prefs.statssec[1],
                                                           prefs.statssec[3],
                                                           lo_rej,
                                                           hi_rej,
                                                           lo_rej);

        if (fabs(scalemedian[i] - lo_rej) < 1.0e-05) {
          e_error("image [%s] has bad median %d",
                  prefs.flatscaleimage[i], lo_rej);
        }

        e_comment(2, "Image %s has median %f", prefs.flatscaleimage[i],
                  scalemedian[i]);

        if (scalemedian[i] > maxflatscale) {
          maxflatscale = scalemedian[i];
        }
        image_del(flatscale);
      }
    }

    e_comment(2, "dividing flat %s by %f", prefs.flatimage,
              maxflatscale);

    for (j = 0; j < (flat->lx * flat->ly); j++) {
      if (fabs(flat->data[j]) > prefs.flatthreshhold) {
        flat->data[j] /= maxflatscale;
      } else {
        flat->data[j] = 0.0;
      }
    }
  }

  if (prefs.fringeflag == 1) {
    e_comment(1, "-> %d.%d: Load Fringe Frame %s", part, 
              subpart++, prefs.fringeimage);

    /* load fringe frame */
    fringe = image_loadext(prefs.fringeimage, 0, 0, 0);

    if (fringe == NULL) {
      e_error("cannot load fringe frame [%s]: aborting", prefs.fringeimage);
      return(-1);
    }

    if (prefs.fringescaleflag == 1) {
      /* load fringe reference frame  for estimating
         fringe scaling
      */
      e_comment(1, "-> using rescaling of fringe images!");

      fringeref = image_loadext(prefs.fringerefimage, 0, 0, 0);

      if (fringeref == NULL) {
        e_error("cannot load fringe reference frame [%s]: aborting", 
                prefs.fringeimage);
        return(-1);
      }

      if ((fringeref->lx != fringe->lx) || (fringeref->ly != fringe->ly)) {
        e_error("fringe ref. frame and fringe frame have incompatible sizes: aborting");
        e_error("fringe ref. frame is [%d %d]", fringeref->lx, fringeref->ly);
        e_error("fringe frame is [%d %d]", fringe->lx, fringe->ly);
        image_del(fringeref);
        return(-1);
      }

      /* estimate median for reference frame */
      fringerefmedian = image_getmedian_vig(fringeref,
                                            prefs.statssec[0],
                                            prefs.statssec[2],
                                            prefs.statssec[1],
                                            prefs.statssec[3]);

      if (fabs(fringerefmedian) < 1e-4) {
        e_error("zero value for median of fringe reference frame: aborting");
        return(-1);
      }

      image_del(fringeref);
    } else {
      e_comment(1, "-> You are not using rescaling of fringe images!");
    }
  }

  e_comment(0, "-> Part %d: performing run processing", ++part);
  subpart = 1;

  /* now do the preprocessing for all the runs */
  for (run = 0; run <= nruns; run++) {
    /* load input images from the current run into a cube */
    nimarun = MIN(prefs.maximages, prefs.ninname - prefs.maximages * run);

    if (nimarun > 0) {
      e_comment(1, "-> %d.%d: loading input images for run %d", part, 
                subpart++, (run + 1));
      i_cube = cube_load_strings(&prefs.inname[run * prefs.maximages], nimarun);

      if (i_cube == NULL) {
        e_error("loading input images", "");
        return(-1);
      }

      /* perform overscan correction */
      if (prefs.overscanflag == 1) {
        e_comment(1, "-> %d.%d: performing overscan correction", part, 
                  subpart++);

        for (i = 0; i < nimarun; i++) {
          corrected = wfi_myoverscan_correction(i_cube->plane[i],
                                                prefs.overscanreg,
                                                prefs.overscanreject);

          if (corrected == NULL) {
            e_error("overscan correction failed for frame %d", i + 1);
            cube_del(i_cube);
            return(-1);
          }
        }
      }

      /* trim images */
      if (prefs.trimflag == 1) {
        e_comment(1, "-> %d.%d: trimming images", part, subpart++);

        if ((prefs.trimreg[0] >= prefs.trimreg[1]) ||
            (prefs.trimreg[2] >= prefs.trimreg[3])) {
          e_error("in crop region definition");
          cube_del(i_cube);
          return(-1);
        }

        for (i = 0; i < nimarun; i++) {
          corrected = image_getvig(i_cube->plane[i],
                                   prefs.trimreg[0],
                                   prefs.trimreg[2],
                                   prefs.trimreg[1],
                                   prefs.trimreg[3]);

          if (corrected == NULL) {
            e_error("trimming failed for frame %d", i + 1);
            cube_del(i_cube);
            return(-1);
          }

          image_del(i_cube->plane[i]);
          i_cube->plane[i] = corrected;
        }
        /* Reset global cube size */
        i_cube->lx = i_cube->plane[0]->lx;
        i_cube->ly = i_cube->plane[0]->ly;
      }

      /* perform biassubtraction */
      if (prefs.biasflag == 1) {
        e_comment(1, "-> %d.%d: Bias subtraction", part, subpart++);

        if ((bias->lx != i_cube->lx) || (bias->ly != i_cube->ly)) {
          e_error("bias and cube have incompatible sizes: aborting");
          e_error("bias is [%d %d]", bias->lx, bias->ly);
          e_error("cube is [%d %d]", i_cube->lx, i_cube->ly);
          cube_del(i_cube);
          image_del(bias);
          return(-1);
        }
        cube_sub_im(i_cube, bias);
      }

      /* perform flatfielding */
      if (prefs.flatflag == 1) {
        e_comment(1, "-> %d.%d: Flatfield correction", part, subpart++);

        if ((flat->lx != i_cube->lx) || (flat->ly != i_cube->ly)) {
          e_error("flat and cube have incompatible sizes: aborting");
          e_error("flat is [%d %d]", flat->lx, flat->ly);
          e_error("cube is [%d %d]", i_cube->lx, i_cube->ly);
          cube_del(i_cube);
          image_del(flat);
          return(-1);
        }

        cube_div_im(i_cube, flat);
      }

      /* perform fringe correction */
      if (prefs.fringeflag == 1) {
        e_comment(0, "-> %d.%d: Fringe correction", part, subpart++);

        if ((fringe->lx != i_cube->lx) || (fringe->ly != i_cube->ly)) {
          e_error("fringe frame and cube have incompatible sizes: aborting");
          e_error("fringe frame is [%d %d]", fringe->lx, fringe->ly);
          e_error("cube is [%d %d]", i_cube->lx, i_cube->ly);
          cube_del(i_cube);
          image_del(fringe);
          return(-1);
        }

        /* if necessary, estimate fringe scaling factors for all science
           frames ans ubtract (scaled) fringeframe:
        */
        for (i = 0; i < nimarun; i++) {
          if (prefs.fringescaleflag == 1) {            
            mediantmp = image_getmedian_vig(i_cube->plane[i],
                                            prefs.statssec[0],
                                            prefs.statssec[2],
                                            prefs.statssec[1],
                                            prefs.statssec[3]);
            fringescaling[i] = mediantmp / fringerefmedian;
          } else {
            fringescaling[i] = 1.0;
          }

          e_comment(1, "Fringe scaling for Image %s: %f", prefs.inname[i],
                      fringescaling[i]);

          for (j = 0; j < (i_cube->plane[i]->lx * i_cube->plane[i]->ly); j++) {
            i_cube->plane[i]->data[j] -=
              (fringe->data[j] * fringescaling[i]);
          }
        }
      } /* prefs.fringeflag */

      /* save result cube to disk in form of single
         images
       */
      if (prefs.outputpreprocflag == 1) {
        e_comment(1, "-> %d.%d: saving preprocessed frames", part, subpart++);

        for (i = 0; i < i_cube->np; i++) {
          /* Build output name */
          sprintf(tmpstr, prefs.outputpreprocsuffix, i);
          sprintf(name_o, "%s/%s%s", prefs.outputdirectory,
                  get_rootname(
                    get_basename(prefs.inname[i + run * prefs.maximages])),
                  tmpstr);
          e_comment(1, "saving [%s]", name_o);

          /* read header from input image */
          fh = qfits_header_read(prefs.inname[i + run * prefs.maximages]);

          if (fh == NULL) {
            e_error("reading FITS header from [%s]", 
                    prefs.inname[i + run * prefs.maximages]);
            return(-1);
          }

          /* collect information about the combination process
             in 'HISTORY' keywords
           */
          hs = history_new();

          if (hs == NULL) {
            e_error("cannot create history object");
            return(-1);
          }

          history_add(hs, "");
          history_add(hs, "preprocess: %s", __PIPEVERS__);
          history_add(hs, "preprocess: called at %s", get_datetime_iso8601());

          if (prefs.overscanflag == 1) {
            history_add(hs, "image has been overscan corrected");
            history_add(hs, "overscan region: %d %d",
                        prefs.overscanreg[0], prefs.overscanreg[1]);
          }

          if (prefs.trimflag == 1) {
            history_add(hs, "image has been trimmed");
            history_add(hs, "trim section: Xmin = %d; Xmax = %d; Ymin = %d; Ymax = %d",
                        prefs.trimreg[0], prefs.trimreg[1],
                        prefs.trimreg[2], prefs.trimreg[3]);
          }

          if (prefs.biasflag == 1) {
            history_add(hs, "image has been bias subtracted");
            history_add(hs, "Bias frame: %s", get_basename(prefs.biasimage));
          }

          if (prefs.flatflag == 1) {
            history_add(hs, "image has been flat fielded");
            history_add(hs, "Flat frame: %s", get_basename(prefs.flatimage));

            if (prefs.flatscaleflag == 1) {
              history_add(hs, "The flatfield has been scaled to %f", 
                          maxflatscale);

              for (j = 0; j < prefs.nflatscaleimage; j++) {
                history_add(hs, "%s %f",
                            get_basename(prefs.flatscaleimage[j]), 
                                         scalemedian[j]);
              }
            } else {
              history_add(hs, "The flatfield has been normalised");
            }
          }

          if (prefs.fringeflag == 1) {
            history_add(hs, "image has been fringe corrected");
            history_add(hs, "Fringe frame: %s; scaling %f",
                        get_basename(prefs.fringeimage),
                        fringescaling[i]);
            history_add(hs, "The scaling was determined w.r.t.: %s",
                        get_basename(prefs.fringerefimage));
          }

          history_addfits(hs, fh);

          /* and finally save */
          if (prefs.outputbitpix == -32) {
            image_save_fits_hdrdump(i_cube->plane[i],
                                    name_o,
                                    fh,
                                    BPP_IEEE_FLOAT);
          }
          if (prefs.outputbitpix == 16) {
            image_save_fits_hdrdump(i_cube->plane[i],
                                    name_o,
                                    fh,
                                    BPP_16_SIGNED);
          }

          qfits_header_destroy(fh);
          history_del(hs);
        }
      }

      /* free memory from the current runs cube; do not do this
         if only one run is involved as we may then still coadd
         the images */
      if (nruns > 0) {
        cube_del(i_cube);
      }
    }
  }

  /* free all calibration frames */
  if (prefs.biasflag == 1) {
    image_del(bias);
  }

  if (prefs.flatflag == 1) {
    image_del(flat);
  }

  if (prefs.fringeflag == 1) {
    image_del(fringe);
  }

  /* combine cube; this is only done if there was
     only ONE run */
  if (prefs.combineflag == 1) {
    if (prefs.combinerescaleflag == 0) {
      e_comment(0, "-> Part %d: stacking cube without rescaling",
                ++part);

      /* check whether the coadded image will be
         useful
       */

      if (i_cube->np - prefs.combinereject[0] - prefs.combinereject[1] < 3) {
        e_warning("coadded image will have the default 0 value everywhere !!!");
        e_warning("Try to reject less low/high values");
      }

      meanmedian = 0.0;       /* default value for combinations without
                                 rescaling */
      stacked    = cube_avg_median_minmax_rej(i_cube,
                                              prefs.combineminthresh,
                                              prefs.combinemaxthresh,
                                              prefs.combinereject[0],
                                              prefs.combinereject[1],
                                              meanmedian);
    } else {
      e_comment(0, "-> Part %d: stacking cube with rescaling",
                ++part);

      /* determine median of all images and rescale to the
         mean of the medians
       */
      e_comment(1, "determine median of input frames:");

      meanmedian = 0.0;
      for (i = 0; i < prefs.ninname; i++) {
        combinemedian[i] = image_getmedian_vig_lowhighreject(i_cube->plane[i],
                                                             prefs.statssec[0],
                                                             prefs.statssec[2],
                                                             prefs.statssec[1],
                                                             prefs.statssec[3],
                                                             lo_rej,
                                                             hi_rej,
                                                             lo_rej);

        if (fabs(combinemedian[i] - lo_rej) < 1.0e-05) {
          e_error("image [%s] has bad median %d",
                  prefs.inname[i], lo_rej);
        }

        e_comment(1, "Image %s: %f", prefs.inname[i],
                  combinemedian[i]);

        if (fabs(combinemedian[i]) < 1e-4) {
          e_error("zero value for median: aborting");
          return(-1);
        }

        meanmedian += combinemedian[i];
      }

      meanmedian /= (pixelvalue)prefs.ninname;

      e_comment(1, "mean of medians; %f", meanmedian);

      for (i = 0; i < prefs.ninname; i++) {
        combinescaling = (meanmedian / combinemedian[i]);

        for (j = 0; j < (i_cube->plane[i]->lx * i_cube->plane[i]->ly); j++) {
          i_cube->plane[i]->data[j] *= combinescaling;
        }
      }

      /* check whether the coadded image will be
         useful
       */
      if (i_cube->np - prefs.combinereject[0] - prefs.combinereject[1] < 3) {
        e_warning("coadded image will have the default value %f everywhere !!! Try to reject less low/high values",
                  meanmedian);
      }

      stacked = cube_avg_median_minmax_rej(i_cube,
                                           prefs.combineminthresh,
                                           prefs.combinemaxthresh,
                                           prefs.combinereject[0],
                                           prefs.combinereject[1],
                                           meanmedian);
    }

    if (stacked == NULL) {
      e_error("in cube combination: aborting");
      return(-1);
    }

    /* create image header for stacked frame */
    fh_stack = qfits_header_default();
    /* BITPIX */
    if (prefs.combinebitpix == -32) {
      sprintf(tmpstr, "%d", BPP_IEEE_FLOAT);
    }
    if (prefs.combinebitpix == 16) {
      sprintf(tmpstr, "%d", BPP_16_SIGNED);
    }
    qfits_header_add(fh_stack, "BITPIX", tmpstr, "Bits per pixel", NULL);
    /* NAXIS* */
    qfits_header_add(fh_stack, "NAXIS", "2", "Number of axes", NULL);
    sprintf(tmpstr, "%d", stacked->lx);
    qfits_header_add(fh_stack, "NAXIS1", tmpstr, "size in X", NULL);
    sprintf(tmpstr, "%d", stacked->ly);
    qfits_header_add(fh_stack, "NAXIS2", tmpstr, "size in Y", NULL);

    /* collect information about the combination process
       in 'HISTORY' keywords
     */
    hs = history_new();

    if (hs == NULL) {
      e_error("cannot create history object");
      return(-1);
    }

    history_add(hs, " ");
    history_add(hs, "preprocess: %s", __PIPEVERS__);
    history_add(hs, "preprocess: %s", get_datetime_iso8601());
    history_add(hs, "preprocess: median of %d images", prefs.ninname);

    history_add(hs, "preprocess: min and max threshhold used: %f %f",
                prefs.combineminthresh, prefs.combinemaxthresh);
    history_add(hs, 
                "preprocess: low and high number of rejected  pixels: %d %d",
                prefs.combinereject[0], prefs.combinereject[1]);

    if (prefs.combinerescaleflag == 1) {
      history_add(hs, "preprocess: combination with rescaling to median %f", 
                  meanmedian);

      for (i = 0; i < prefs.ninname; i++) {
        history_add(hs, "preprocess: %s %.2f",
                    get_basename(prefs.inname[i]),
                    combinemedian[i]);
      }
    } else {
      history_add(hs, "preprocess: combination without rescaling");

      for (i = 0; i < prefs.ninname; i++) {
        history_add(hs, "preprocess: %s",
                    get_basename(prefs.inname[i]));
      }
    }

    history_addfits(hs, fh_stack);

    if (prefs.combinebitpix == -32) {
      image_save_fits_hdrdump(stacked, prefs.combineoutput,
                              fh_stack, BPP_IEEE_FLOAT);
    }
    if (prefs.combinebitpix == 16) {
      image_save_fits_hdrdump(stacked, prefs.combineoutput,
                              fh_stack, BPP_16_SIGNED);
    }
    qfits_header_destroy(fh_stack);
    history_del(hs);
    image_del(stacked);
  }

  /* free memory and bye */
  free(prefs.inname);
  free(combinemedian);
  free(scalemedian);
  free(fringescaling);

  if (nruns == 0) {
    cube_del(i_cube);
  }

  if (debug_active()) {
    xmemory_status();
  }
  return(0);
}

/*
 * This function only gives the usage for the program
 */

void printUsage()
{
  fprintf(stdout, "PROGRAMNAME\n");
  fprintf(stdout, "        preprocess - Mosaic camera preprocessing software\n");
  fprintf(stdout, "\n");
  fprintf(stdout, "SYNOPSIS\n");
  fprintf(stdout, "        preprocess images [options]\n");
  fprintf(stdout, "                   -h    # print this help and leave\n");
  fprintf(stdout, "                   -d    # dump standard configuration file and leave\n");
  fprintf(stdout, "                   -c    # name of the configuration file\n");
  fprintf(stdout, "\n");
  fprintf(stdout, "DESCRIPTION \n");
  fprintf(stdout, "        The program performs all steps necessary to take the\n");
  fprintf(stdout, "        instrumental signature out of optical astronomical CCD\n");
  fprintf(stdout, "        data. Given a set of input images, the program can perform the\n");
  fprintf(stdout, "        following operations, or any specified subset, on the data:\n");
  fprintf(stdout, "        \n");
  fprintf(stdout, "        - overscan correction:\n");
  fprintf(stdout, "          From each line, the pixels in the overscan region\n");
  fprintf(stdout, "          (OVERSCAN_REGION) are collected, the lowest and highest\n");
  fprintf(stdout, "          values specified in OVERSCAN_REJECT rejected and the mean\n");
  fprintf(stdout, "          estimated. That mean is subtracted from all pixels in that\n");
  fprintf(stdout, "          line.\n");
  fprintf(stdout, "        \n");
  fprintf(stdout, "        - Trimming:\n");
  fprintf(stdout, "          all input images are trimmed to the section specified in\n");
  fprintf(stdout, "          TRIM_REGION\n");
  fprintf(stdout, "        \n");
  fprintf(stdout, "        - Bias subtraction:\n");
  fprintf(stdout, "          The image specified with BIAS_IMAGE is subtracted from all\n");
  fprintf(stdout, "          input images.\n");
  fprintf(stdout, "        \n");
  fprintf(stdout, "        - Flat fielding:\n");
  fprintf(stdout, "          All images are divided by the image specified in\n");
  fprintf(stdout, "          FLAT_IMAGE. If scaling (FLAT_SCALE) is asked for, the\n");
  fprintf(stdout, "          flatfield is normalised to the highest median of the images\n");
  fprintf(stdout, "          specified in FLAT_SCALEIMAGE.  Otherwise the flat is\n");
  fprintf(stdout, "          normlised to a mean of unity.\n");
  fprintf(stdout, "        \n");
  fprintf(stdout, "        - Fringe correction:\n");
  fprintf(stdout, "          A fringe pattern given with FRINGE_IMAGE is subtracted from\n");
  fprintf(stdout, "          all images. The fringe pattern is scaled with\n");
  fprintf(stdout, "          median(FRINGE_REFIMAGE)/median(image), i.e. we assume that\n");
  fprintf(stdout, "          the intensity of the fringes scales with the sky background.\n");
  fprintf(stdout, "          This scaling can be turned off with the FRINGE_SCALE option.\n");
  fprintf(stdout, "        \n");
  fprintf(stdout, "        - Result output:\n");
  fprintf(stdout, "          All preprocessed images are saved to disk in a directory\n");
  fprintf(stdout, "          given in OUTPUT_DIR. The ending '.fits' of each input image\n");
  fprintf(stdout, "          is replaced by OUTPUT_SUFFIX.\n");
  fprintf(stdout, "        \n");
  fprintf(stdout, "        - Image stacking:\n");
  fprintf(stdout, "          THIS IS ONLY AVAILABLE IF ALL INPUT IMAGES CAN\n");
  fprintf(stdout, "          SIMULTANEOUSLY BE LOADED INTO VIRTUAL MEMORY; SEE BELOW !!!!\n");
  fprintf(stdout, "          all images are stacked. At each image position, the\n");
  fprintf(stdout, "          individual pixel values are collected. We reject all pixels\n");
  fprintf(stdout, "          being smaller than COMBINE_MINVAL and larger than\n");
  fprintf(stdout, "          COMBINE_MAXVAL. From the remaining pixels we further do not\n");
  fprintf(stdout, "          consider the COMBINE_LOREJ lowest and COMBINE_HIREJ highest\n");
  fprintf(stdout, "          values. From the rest we estimate the median. If rescaling\n");
  fprintf(stdout, "          (COMBINE_RESCALE) is demanded we rescale all images to the\n");
  fprintf(stdout, "          mean of the medians from all input data prior to coaddition.\n");
  fprintf(stdout, "          If at some stage less than 3 pixels are left, the result is\n");
  fprintf(stdout, "          the mean of the medians of input images in the rescale case\n");
  fprintf(stdout, "          and 0 otherwise. If less than 3 images are supplied in the\n");
  fprintf(stdout, "          beginning, no stacking is performed. The result is saved to\n");
  fprintf(stdout, "          COMBINE_OUTPUT.\n");
  fprintf(stdout, "        \n");
  fprintf(stdout, "        For all operations requiring statistics on the input frames,\n");
  fprintf(stdout, "        the image region specified in STATSSEC is used.\n");
  fprintf(stdout, "        \n");
  fprintf(stdout, "        The program loads consecutively frames in chunks of MAXIMAGES\n");
  fprintf(stdout, "        into the memory. Set it to a value so that MAXIMAGES frames\n");
  fprintf(stdout, "        fit into the RAM of your machine. All processes (except\n");
  fprintf(stdout, "        stacking) can naturally be split in this way. Hence,\n");
  fprintf(stdout, "        coaddition is NOT available if more than MAXIMAGES frames have\n");
  fprintf(stdout, "        to be processed.\n");
  fprintf(stdout, "\n");
  fprintf(stdout, "EXAMPLES:\n");
  fprintf(stdout, "        - creation of a master bias frame:\n");
  fprintf(stdout, "          the creation of a master bias frame usually consists\n");
  fprintf(stdout, "          of overscan correction, trimming and combination\n");
  fprintf(stdout, "          without rescaling. This could be achieved with:\n");
  fprintf(stdout, "        \n");
  fprintf(stdout, "          preprocess *.fits -OVERSCAN Y -TRIM Y -COMBINE Y\n");
  fprintf(stdout, "                            -COMBINE_RESCALE N \n");
  fprintf(stdout, "                            -COMBINE_OUTPUT BIAS.fits\n");
  fprintf(stdout, "        \n");
  fprintf(stdout, "          The default values for overscan-, statistic region\n");
  fprintf(stdout, "          and rejection values in the coaddition process would\n");
  fprintf(stdout, "          have to be updated if necessary.\n");
  fprintf(stdout, "        \n");
  fprintf(stdout, "\n");
  fprintf(stdout, "AUTHOR\n");
  fprintf(stdout, "        Thomas Erben: terben@astro.uni-bonn.de\n");
  fprintf(stdout, "\n");
  fprintf(stdout, "        The code is largely based on the eclipse library (see\n");
  fprintf(stdout, "        http://www.eso.org/eclipse) and on sources from Emmanuel\n");
  fprintf(stdout, "        Bertin for the handling of configuration options.\n");
}
