#!/usr/bin/env bash

# Main installation script for the THELI pipeline

# IMPORTANT:
#
# The script checks machine type, updates pipeline
# scripts (progs.ini; pipe_env ...)
# and compiles pipeline sources
# the script needs the source file 'endian.c' in the
# directory it is called. Modify the paths listed below
# before running this script.

#
# Modes of the script
# 
# ./install.sh -m "TEST"
# tests the machine type and gives information whether
# you can compile sources without problems or whether
# you have to create new Makefiles (see the documentation
# for more details)
#
# ./install.sh -m "ALL"
# updates paths in all pipeline scripts and compiles
# pipeline sources
#
# ./install.sh -m "UPDATE"
# only updates paths in pipeline scripts
#
# ./install.sh -m "COMPILE"
# only compiles pipeline sources

#
# HISTORY COMMENTS:
# 
# 09.03.2006:
# - I removed the utilities scripts subdirectory from
#   the script updates. It contains no files needed in
#   the official THELI version.
# - I included MAC/Darwin systems to the known architectures.
#
# 14.03.2006:
# I changed the installation scheme for ASTROMETRIX: instead
# of overriding the original script files with modified ones,
# I keep the modified, renamed, ones in the pipeline structure
# and add the corresponding directory the PERL5LIB variable. 
# This allows us to directly call the modified ASTROMETRIX files 
# with the perl interpreter.
#
# 27.04.2006:
# I adapted the location of 'global_vers.h' (it is now in
# ../ldactools/library and no longer in ../library)
#
# 01.05.2006:
# The pipeline version for the external version is
# 'THELI Pipeline Version: XXXX'. This allows a distinction
# to the internal sources.
#
# 15.05.2006:
# - I included a simple check for the presence of necessary, external 
#   programs defined in this script.
# - The script now prints some description when being invoked.
#
# 17.05.2006:
# - I changed slightly the welcome message
# - increase of the minor pipeline version number. 
# - I improved messages during the installation process.
# - The handling of the test for the availabbility of external
#   UNIX programs is more stable to account for architecture dependent 
#   behaviours of the 'which' program.
#
# 14.06.2006:
# - The availability of the external 'sm' program is checked via
#   a 'which' command now. 
# - Increase of minor pipeline version number.
#
# 30.08.2006:
# Increase of minor pipeline version number.
#
# 05.11.2006:
# - If the program 'sm' is not present, the variable P_SM
#   is set to an empty string. Without this the value could
#   be quite undefined. This avoids potential problems
#   in the 'sed' command which modifies the progs.ini file.
# - In the 'sed' command which modifies the progs.ini file 
#   I removed lines refering to IRAF.
#
# 06.11.2006:
# The version number goes to 0.12.0
#
# 18.09.2007:
# I included code for the possibility to use scamp within
# THELI. The script documentation was updated accordingly.
#
# 03.12.2007:
# I introduced Darwin_PPC and Darwin_INTEL systems; the old
# 'Darwin' was removed. This was necessary as the Intel and PPC
# MACS come with different byte swap for floating point numbers.
#
# 06.07.2008:
# Cosmetic changes to avoid error messages during install
#
# 21.12.2008:
# - The interpreters for python and perl are defined now
#   also via variables. The main problem was that 'alias'
#   deinitions were not taken into account in pipeline
#   script calls leading to problems when the script 
#   interpreters were privately installed.
# - I updated the documentation
#
# 09.02.2010:
# Improvements to the verification of installation requirements.
#
# 15.03.2010:
# Bug Fix: Configuration files for WFI were not copied correctly
# to the 'MACHINE' directory
#
# 30.03.2010:
# I included the perl script 'stellar_locus.pl' which needed 
# modification in the updatescripts part.
#
# 31.08.2010:
# Version number goes to 1.2.1
#
# 16.09.2010:
# Version number goes to 1.2.2 (internal production version)
#
# 16.09.2010:
# Version number goes to 1.3.0 (new official release)
#
# 08.10.2010:
# Version number goes to 1.3.1 (internal production version)
#
# 26.11.2010:
# - MEGAPRIME files are now included in the distribution
# - Version number goes to 1.4.0 (internal production version);
#   new swarp version and associated changes
#
# 17.01.2011:
# version number goes to 1.4.1 (internal production version)
#
# 24.05.2011:
# version number goes to 1.5.0 (release version; several important bug fixes)
#
# 13.06.2011:
# Version number goes to 1.5.1; I included Makefiles for Darwin_INTEL 64
# architectures.
#
# 27.07.2011:
# Version number goes to 1.5.2 (minor bug fixes)
#
# 05.09.2011:
# Version number goes to 1.5.3 (minor bug fixes / enhancements)
#
# 26.01.2012:
# Version number goes to 1.6.0 (bug fixes / enhancements)
#
# 31.01.2012:
# Version number goes to 1.6.1 (bug fixes / enhancements)

##
## stuff that has to be modified by the user
##

# needed directories (absolutely necessary)
#
# PD is the main directory with the pipeline sources,
# i.e. the .../theli_xxxx directory. PD has to contain the
# complete, absolute path. 
#
# Make a 'good guess' for PD. DO NOT EDIT !!
DIR=`pwd`
cd ..
PD=`pwd`
cd ${DIR}

# needed programs/scripts not included in the
# pipeline distribution (absolutely necessary)
#
# The path for the bash shell (you need version
# 3.0 or higher)
P_BASH=`which bash`
#
# The path to the GNU awk program (awk is NOT sufficient)
P_GAWK=`which gawk`
#
# The path of the UNIX find program (the '-maxdepth' option
# has to be supported; use for instance the GNU find !!)
P_FIND=`which find`
#
# The path of the UNIX sort program (the '-g' option
# has to be supported; use for instance the GNU sort !!)
P_SORT=`which sort`
#
# The path to the perl and python interpreters called
# by pipeline scripts:
P_PYTHON=`which python`
P_PERL=`which perl`

#
# the following programs are not absolutely
# necessary (see the manual to learn what restrictions
# you have in this case)
#
#
# Path to the Super Mongo (sm) program
P_SM=`which sm`
#
# Path to the scamp program. The current THELI
# pipeline supports V1.7.0 of that program
#
P_SCAMP=`which scamp`
#
# Path to the aclient program. It is mandatory
# if scamp is used within THELI
P_ACLIENT=`which aclient`


# Temporary directory for small files created during the
# coaddition process. Giving '.' will put all these files
# into the reduction directory.
TEMPDIR=.

#
#
# C-compiler used in this script
# (this affects ONLY this script; not the
# C-compiler used in the compilation of the sources)
CC=gcc

#
# known architectures for source compilation (see the documentation
# for further information)
KNOWNARCH="AIX Darwin_INTEL Darwin_INTEL_64 Darwin_PPC HP-UX IRIX64 Linux Linux_64 OSF1 SunOS"

##
## here the stuff that should ever be edited by a regular user ends!!!
##


#
# Pipeline version
PIPEVERS="THELI Pipeline Version: 1.6.1"

#
# subprogram definitions
#

#
# by default the architecture is not known !!
KNOWN=0
MACHINE=""

function updatescripts() {
  DIR=`pwd`
  cd $1
  #
  # deal with the bash shell
  # scripts where we substitute the
  # first line by the actual location
  # of the bash shell
  SCRIPTFILES=`find . -name \*.sh.raw`

  for SCRIPT in ${SCRIPTFILES}
  do
    BASE=`basename ${SCRIPT} .sh.raw`
    sed -e 's!/bin/bash!'${P_BASH}'!' \
           ${SCRIPT} > ${PD}/scripts/${MACHINE}/${BASE}.sh	  
    chmod a+x ${PD}/scripts/${MACHINE}/${BASE}.sh
  done

  cd ${DIR}
}

#
# test the machine type
# run also sysconf from eclipse
# check the presence of several UNIX tools and
# optional programs
#
function testmachine() {
  # first test if the endian program
  # is alreay available; if not, compile
  # it!
  if [ ! -f endian.c ]; then
    echo "need the source file 'endian.c'"
    exit 1
  else
    ${CC:?} -o endian endian.c	 
  fi

  echo "---------------------------------------------------"
  tput bold
  echo "Detect your architecture:"
  tput sgr0
  echo ""

  MACHINE=`uname`
  #
  # test for LINUXALPHA architecture if
  # the machine is Linux
  if [ "${MACHINE}" = "Linux" ]; then
    LONG=`./endian | ${P_GAWK:?} '($1=="Sizes:") {print substr($5,3,1)}'`

    if [ "${LONG}" = "8" ]; then
      MACHINE="Linux_64"
    fi
  fi

  #
  # test for the different MAC/Darwin systems
  # (PPC or Intel)
  if [ "${MACHINE}" = "Darwin" ]; then
    INTEL=`./endian | ${P_GAWK:?} '($2=="machine") {print $5}'`

    MACHINE="Darwin_PPC"
    if [ "${INTEL}" = "Intel" ]; then
      MACHINE="Darwin_INTEL"
    fi

    LONG=`./endian | ${P_GAWK:?} '($1=="Sizes:") {print substr($5,3,1)}'`

    if [ "${LONG}" = "8" ]; then
      MACHINE="${MACHINE}_64"
    fi
  fi

  #
  # test for known architecture
  for arch in ${KNOWNARCH}
  do
    if [ "${MACHINE}" = "${arch}" ]; then
	KNOWN=1
    fi 
  done    

  if [ "${KNOWN}" = "0" ]; then
    tput bold
    tput setf 4
    echo "Your machinetype \"${MACHINE}\" has not yet been tested !!"
    echo "Please refer to the manual how to proceed"
    echo ""
    ENDIAN=`./endian |  ${P_GAWK:?}'($1 == "this") {print $5}'`

    if [ "${ENDIAN}" = "Intel" ]; then
       echo "You need byte swap !!! (-DBSWAP flag)"
    else
       echo "You do not need byte swap !!!"
    fi

    LONG=`./endian | ${P_GAWK:?}  '($1=="Sizes:") {print substr($5,3,1)}'`

    if [ "${LONG}" = "8" ]; then
       echo "Long is represented by 8 bytes (-DINT64 flag)"
    else
       echo "Long is represented by 4 bytes"
    fi
    tput sgr0

    exit 1
  else
    echo "Your are running a \"${MACHINE}\" machine."
    echo "The compilation of THELI sources should work without problems"
    echo ""
  fi
  #
  # running eclipse's system configuration:
  #

  echo "---------------------------------------------------"
  echo ""
  tput bold
  echo "Run eclipse system configuration tool:"
  tput sgr0
  echo ""

  if [ ! -f sysconf.c ]; then
    echo "need the source file 'sysconf.c'"
    exit 1
  else
    ${CC:?} -o sysconf sysconf.c	 
  fi
  ./sysconf

  cp ./config.h ${PD}/eclipsetools/eclipse/include
  mv ./config.h ${PD}/eclipsetools/qfits

  #
  # do a simple check whether the necessary UNIX programs defined above are at 
  # least present and executable. Passing this test DOES NOT ensure that the
  # programs have the required functionality, though.

  echo ""
  echo "---------------------------------------------------"
  echo ""
  tput bold
  echo "Checking the presence of the absolutely necessary UNIX tools:"
  echo "'bash', 'gawk', 'find', 'sort':"
  tput sgr0
  echo ""

  NEEDED_PROGRAMS="P_BASH P_GAWK P_FIND P_SORT"
  OPTIONAL_PROGRAMS="P_SM P_SCAMP P_ACLIENT"
  
  for PROGRAM in ${NEEDED_PROGRAMS}
  do
    # if the 'which' command to obtain program paths fails
    # it outputs several words which lead to an error
    # in the following 'if'. Hence the following TESTPROG
    # definition.
    TESTPROG=`echo ${!PROGRAM} | awk '{print $1}'`
    if [ "A_${TESTPROG}" = "A_" ] || [ ! -x ${TESTPROG} ]; then
      tput bold
      tput setf 4        
      echo "Program defined in ${PROGRAM} is not in your PATH or not executable"
      echo "Please correct this error before continuing !!"
      echo ""
      tput sgr 0

      exit 1
    fi
  done

  echo "'bash', 'gawk', 'find' and 'sort' seem to be present."
  echo "This test does not ensure that they fulfil the criteria"
  echo "documented in 'install.sh'"
  echo ""

  echo "---------------------------------------------------"
  echo ""
  tput bold
  echo "Checking of optional programs:"
  echo "sm, scamp, aclient:"
  tput sgr0
  echo ""

  for PROGRAM in ${OPTIONAL_PROGRAMS}
  do
    TESTPROG=`echo ${!PROGRAM} | awk '{print $1}'`
    if [ "A_${TESTPROG}" = "A_" ] || [ ! -x ${TESTPROG} ]; then
      tput bold
      tput setf 1
      echo "Program defined in ${PROGRAM} is not your PATH or not executable"
      echo "You can still use the THELI pipeline but with limited functionality !!"
      tput sgr0
      eval ${PROGRAM}=""
    else
      echo "Program defined in ${PROGRAM} is present. The test does not ensure"
      echo "that the executable represents the correct program."  
    fi
  done
}

#
# test version numbers of Python modules
function testpythonmodules() {
  echo "---------------------------------------------------"
  ./check_python_modules.sh ${P_PYTHON}
}

#
# create setup script responsible for setting
# the pipeline environment variables:
#
function environment() {
  # just to be sure:
  # first test the machine type if necessary
  if [ "${MACHINE}" = "" ]; then
    testmachine
  fi

  # create the pipe_env scripts
  
  sed -e 's!PIPESOFT=!PIPESOFT='${PD}'!' \
      -e 's!EXT=!EXT='${MACHINE}'!' \
      ${PD}/pipesetup/pipe_env.bash.raw \
      > ${PD}/pipesetup/pipe_env.bash.${MACHINE}
  
  sed -e 's! PIPESOFT! PIPESOFT '${PD}'!' \
      -e 's! EXT! EXT '${MACHINE}'!' \
      ${PD}/pipesetup/pipe_env.csh.raw \
      > ${PD}/pipesetup/pipe_env.csh.${MACHINE}
}

#
# print script help message:
function helpmessage() {
  echo "Install script for the THELI core pipeline"
  echo "=========================================="
  echo " "
  echo "PLEASE READ THE FOLLOWING COMPLETELY BEFORE DOING ANYTHING!"
  echo ""
  echo "This script executes necessary steps to compile the THELI sources and"
  echo "to prepare the THELI reduction scripts for your system. Before"
  echo "continuing, make sure the following preconditions:"
  echo ""
  echo "- make sure that you have the GNU C-compiler (gcc) working on your"
  echo "  system. All gcc V3.X and V4.X should work without problem.  "
  echo ""
  echo "- THELI needs the following external programs for full functionality:"
  echo "  - The UNIX/Linux tools: bash shell (V3.X), 'gawk', 'find', 'sort'."
  echo "    Those should be present on a standard UNIX/Linux/MacOS installation"
  echo "    that is suitable for scientific computing."
  echo ""
  echo "  - Emmanuel Bertins 'scamp' program for astrometric calibration (see"
  echo "    http://www.astromatic.net/software/scamp); that program in turn"
  echo "    needs the cdsclient package which contains the 'aclient' program"
  echo "    (see http://cdsweb.u-strasbg.fr/doc/cdsclient.html)"
  echo ""
  echo "  - The SuperMongo (sm) tool for producing control plots. This program"
  echo "    is non-free, but it is typically available at astronomical"
  echo "    institutes (see also http://www.astro.princeton.edu/~rhl/sm/)"
  echo ""
  echo "  - The Python language interpreseter in version V2.Y with Y>=5. In addition,"
  echo "    the python modules 'numpy' (V1.3.0 or higher), 'matplotlib'"
  echo "    (V0.98.1 or higher) and 'pyfits' (V1.1 or higher) are"
  echo "    needed. Python, and those modules are used for essential parts of"
  echo "    astrometric and photometric calibration."
  echo ""
  echo "  This install script has a testing mode (see below) which tries to"
  echo "  verify these preconditions. PLEASE run this mode and CORRECT ALL"
  echo "  potential problems before further installation."
  echo "  "
  echo "- Besids 'scamp' THELI can use Mario Radovichs ASTROMETRIX for astrometric "
  echo "  calibration. A special THELI version of that software is shipped as a "
  echo "  separate package (WIFIX_Distrib_X.X_theli.tgz; the original sources"
  echo "  from Mario can be found at http://www.na.astro.it/~radovich/wifix.htm)"
  echo "  If you want to use it, install it BEFORE continuing with the installation "
  echo "  of this package."
  echo ""
  echo "The actual THELI compilation and script preparation phases are done with the"
  echo "following calls to this script:"
  echo ""
  echo "- ./install.sh -m 'TEST'"
  echo "  This tests your architecture and reports whether THELI can be"
  echo "  compiled without further preparations. Running this mode should"
  echo "  give an output such as:"
  echo ""
  echo "  Your are running a 'MACHINE' machine."
  echo "  The compilation of THELI sources should work without problems"
  echo "  ."
  echo "  ."
  echo "  ."
  echo ""
  echo "  Please note the value of your MACHINE (probably Linux or Linux_64)."
  echo "  Also correct errors reported at the end of the output."
  echo ""
  echo "- ./install.sh -m 'ALL'"
  echo "  This compiles the THELI sources and adaptes the pipeline scripts to"
  echo "  your system. It is a combination of the COMPILE and UPDATE modes"
  echo "  and represents the complete installation task of this script."
  echo ""
  echo "- ./install.sh -m 'COMPILE'"
  echo "  This compiles THELI sources. After compilation, executables can be found"
  echo "  in .../theli-X.XX.XX/bin/MACHINE where MACHINE is your computer"
  echo "  architecture. "
  echo ""
  echo "- ./install.sh -m 'UPDATE'"
  echo "  This adapts the pipeline scripts to your system. The modified and "
  echo "  ready-to-use scripts can be found in .../theli-X.XX.XX/scripts/MACHINE"
  echo ""
  echo "To finish the setup, add the lines in the file"
  echo ".../theli-X.XX.XX/pipesetup/pipe_env.bash.MACHINE (bash shell) or"
  echo ".../theli-X.XX.XX/pipesetup/pipe_env.csh.MACHINE (C-shell) to your"
  echo "shell startup file ('.bashrc' if you use bash and '.cshrc' for the"
  echo "tc-shell respectively). You find the THELI data processing scripts in"
  echo "the directory .../theli-X.XX.XX/scripts/MACHINE."
  echo ""
}


# read command line argument(s)

if [ $# -lt 2 ]; then
  helpmessage
  exit 1
fi

# 4 MODES available:
# - TEST: the the machine type
# - ALL: do script update and compile sources
# - UPDATE: do script update
# - COMPILE: compile sources

MODE=""
GoOn=0
while [ $GoOn = 0 ]
do
  case $1 in
  -m)
      MODE=${2}
      shift
      shift
      ;;
   *)
      GoOn=1
      ;;
  esac
done

#
# test machine type
#
for mode in ${MODE}
do
  if [ "${mode}" = "TEST" ]; then
    testmachine
    testpythonmodules
  fi    
done

#
# update scripts
#
for mode in ${MODE}
do
  if [ "${mode}" = "UPDATE" ] || [ "${mode}" = "ALL" ]; then

    # first test the machine type if necessary
    if [ "${MACHINE}" = "" ]; then
      testmachine
    fi

    #
    # create machine dependent subdirectory in the
    # scripts directory
    if [ ! -d ${PD}/scripts/${MACHINE} ]; then
      mkdir ${PD}/scripts/${MACHINE}
    fi

    # create setup script for pipeline environment
    # if necessary:
    if [ ! -f ${PD}/pipesetup/pipe_env.bash.${MACHINE} ]; then
      environment
    fi
    
    # now the progs.ini file    
    sed -e 's!PIPESOFT=!PIPESOFT='${PD}'!' \
        -e 's!^PHOTSTANDARDSTARSCAT=!PHOTSTANDARDSTARSCAT='${PHOTSTANDARDSTARCAT}'!' \
        -e 's!^STANDARDSTARSCAT=!STANDARDSTARSCAT='${STANDARDSTARCAT}'!' \
        -e 's!TEMPDIR=.!TEMPDIR='${TEMPDIR}'!' \
        -e 's!BIN=${PIPESOFT}/bin/!BIN=${PIPESOFT}/bin/'${MACHINE}'/!' \
        -e 's!UTILSCRIPTS=${PIPESOFT}/scripts/utilities!UTILSCRIPTS=${PIPESOFT}/scripts/'${MACHINE}'/!' \
        -e 's!P_PYTHON=!P_PYTHON='${P_PYTHON}'!' \
        -e 's!P_PERL=!P_PERL='${P_PERL}'!' \
        -e 's!P_GAWK=!P_GAWK='${P_GAWK}'!' \
        -e 's!P_FIND=!P_FIND='${P_FIND}'!' \
        -e 's!P_SORT=!P_SORT='${P_SORT}'!' \
        -e 's!P_SM=!P_SM='${P_SM}'!' \
        -e 's!P_SCAMP=!P_SCAMP='${P_SCAMP}'!' \
        -e 's!P_ACLIENT=!P_ACLIENT='${P_ACLIENT}'!' \
        -e 's!S_ASTROMETRIX=!S_ASTROMETRIX='${PD}'/scripts/astrometrix/astrom_theli!' \
        -e 's!S_PHOTOMETRIX=!S_PHOTOMETRIX='${PD}'/scripts/astrometrix/photom_theli!' \
        ${PD}/scripts/reduction/progs.ini.raw \
        > ${PD}/scripts/${MACHINE}/progs.ini
    
    # now update all the scripts
    SCRIPTDIRS="${PD}/scripts/reduction ${PD}/scripts/WFI_reduction
                ${PD}/scripts/reduction ${PD}/scripts/CFH12K_reduction
                ${PD}/scripts/MEGAPRIME_reduction
                ${PD}/scripts/parallel  ${PD}/scripts/utilities"
    
    for SCRIPTDIR in ${SCRIPTDIRS}
    do
      updatescripts ${SCRIPTDIR}
      #
      # just copy possible Python scripts
      # (they are called via the env command and
      # hence we do not need to hardcode the location
      # of the Python executable):
      PERLPYFILES=`find ${SCRIPTDIR} \( -name \*.pl -o -name \*.py \)`

      for perlpyfile in ${PERLPYFILES}
      do
	BASE=`basename ${perlpyfile}`
        cp ${perlpyfile} ${PD}/scripts/${MACHINE}
	chmod a+x ${PD}/scripts/${MACHINE}/${BASE}
      done
      #
      # copy include files
      INCLUDE=`find ${SCRIPTDIR} -name \*.include`

      for include in ${INCLUDE}
      do
        cp ${include} ${PD}/scripts/${MACHINE}
      done
    done

    # copy files from the instrument specific directories
    # to the machine directories
    cp ${PD}/scripts/WFI_reduction/WFI* ${PD}/scripts/${MACHINE}
    cp ${PD}/scripts/CFH12K_reduction/CFH12K* ${PD}/scripts/${MACHINE}
    cp ${PD}/scripts/CFH12K_reduction/cfh12k* ${PD}/scripts/${MACHINE}
    cp ${PD}/conf/WFI_reduction/WFI* ${PD}/scripts/${MACHINE}
    cp ${PD}/scripts/MEGAPRIME_reduction/MEGAPRIME* ${PD}/scripts/${MACHINE}
    cp ${PD}/conf/MEGAPRIME_reduction/MEGAPRIME* \
       ${PD}/conf/MEGAPRIME_reduction/megacam.ret ${PD}/scripts/${MACHINE}

    chmod a+x ${PD}/pipesetup/test_dir
    chmod a+x ${PD}/pipesetup/vraag

    # now deal with the perl scripts from ASTROMETRIX:
    sed -e "s!GaBoDSVersion=!GaBoDSVersion=\"${PIPEVERS}\";!" \
      ${PD}/scripts/astrometrix/src/photom_theli.raw >\
      ${PD}/scripts/astrometrix/photom_theli
    chmod +x ${PD}/scripts/astrometrix/photom_theli
    sed -e "s!GaBoDSVersion=!GaBoDSVersion=\"${PIPEVERS}\";!" \
      ${PD}/scripts/astrometrix/src/astrom_theli.raw >\
      ${PD}/scripts/astrometrix/astrom_theli
    chmod +x ${PD}/scripts/astrometrix/astrom_theli
    sed -e "s!GaBoDSVersion=!GaBoDSVersion=\"${PIPEVERS}\";!" \
      ${PD}/scripts/astrometrix/src/nldacsub_theli.pm.raw >\
      ${PD}/scripts/astrometrix/nldacsub_theli.pm
    cp ${PD}/scripts/astrometrix/src/AstSubs_theli.pm.raw \
       ${PD}/scripts/astrometrix/AstSubs_theli.pm
    cp ${PD}/scripts/astrometrix/src/AstConf_theli.pm.raw \
       ${PD}/scripts/astrometrix/AstConf_theli.pm
    cp ${PD}/scripts/astrometrix/src/catsubs_theli.pm.raw \
       ${PD}/scripts/astrometrix/catsubs_theli.pm
    cp ${PD}/scripts/astrometrix/src/PhotConf_theli.pm.raw \
       ${PD}/scripts/astrometrix/PhotConf_theli.pm
  fi
done

# compile sources

for mode in ${MODE}
do
  if [ "${mode}" = "COMPILE" ] || [ "${mode}" = "ALL" ]; then
    # do a test first whether the machine type
    # is already known
    if [ "${MACHINE}" = "" ]; then
      testmachine
    fi

    # create setup script for pipeline environment
    # if necessary:
    if [ ! -f ${PD}/pipesetup/pipe_env.bash.${MACHINE} ]; then
      environment
    fi

    # set environment variables necessary for compilation
    . ${PD}/pipesetup/pipe_env.bash.${MACHINE}

    # write the correct global_vers.h file:
    sed -e "s!#define __PIPEVERS__!#define __PIPEVERS__ \"${PIPEVERS}\"!" \
	  ${PD}/pipesetup/global_vers.h.raw >\
	  ${PD}/ldactools/library/global_vers.h

    cd ${PD}/pipesetup/
    make all
    cd ${PD}/pipesetup/
  fi
done









