#!/bin/bash -u

# -----------------------------------------------------------------------
# File Name:           create_scamp_astrom_photom.sh
# Author:              Thomas Erben (terben@astro.uni-bonn.de)
# Last modified on:    27.05.2011
# Description:         Performs astrometric and relative photometric
#                      calibration of THELI sets with scamp V1.4.0-V1.7.0
# -----------------------------------------------------------------------

# Script history information:
# 14.09.2007:
# Project started 
#
# 17.09.2007:
# - I updated the documentation
# - I fixed some bugs concerning the location of
#   resulting files
# - I appended NTHREADS to the scamp command line
#   which uses a multi-processor architecture if available.
#   NTHREADS is set to NPARA.
#
# 24.09.2007:
# - Update of documentation text
# - Bug fix for the location of catalogue files
# - FITS conform output of the RZP header keyword in
#   scamp catalogues
#
# 02.10.2007:
# I made the script a bit more robust by adding more sanity
# checks.
#
# 21.12.2007:
# - I extended the script so that different mosaic configurations
#   within one astrometric set can be handled now. To this end we
#   create an artificial external header containing the keyword
#   MISSCHIP listing the missing chips. This keyword is then added
#   to the FITS cards distinguishing astrometric runs.
# - One explicit call to 'ldactoasc' was replaced by an implicit one.
#
# 30.12.2007:
# - I extended the script to treat data from different image directories
#   simultaneously.
# - Multiple photometric contexts are now handled correctly.
#
# 08.12.2008:
# I updated the script documentation
#
# 11.01.2008:
# The script now makes use of focal plane information if present!
# If an external header with name ${INSTRUMENT}.ahead is present in
# the reduction directory it is interpreted as containing focal plane
# information (first order astrometric information). In this case
# 'scamp' is run with the '-MOSAIC_TYPE FIX_FOCALPLANE' option. 
# Otherwise (no focal plane information available) we use
# '-MOSAIC_TYPE UNCHANGED'.
#
# 16.01.2008:
# I estimate the 'THELI' relative flux scaling zeropoint. Up to
# now I used the scamp value which is, however, scaled to a fixed
# magnitude (scamp config file). This value is not consistent with
# the THELI convention where the flux scale for each image is defined 
# as flxscale=10**(-0.4*relzp)/exptime
#
# 18.01.2008:
# Bug fix in the creation of individual header files; missing catalogs
# were not taken into account properly.
#
# 22.07.2008:
# The directory names where scamp performs its actions carry the name 
# of the used standardstar catalogue now. The same applies for the
# final header directories.
#
# 07.08.2008:
# Bug Fix for the new naming convention of appending the name of the
# reference cat to directory names
#
# 21.12.2008:
# The python interpreter is now called via a variable which is defined
# in 'progs.ini'.
#
# 20.01.2009:
# It was found that CFHTS-DEEP files 'shuffled' in time, when given to
# scamp, lead to better results! Better results means better looking
# astrometry control plots!
#
# 23.02.2009:
# Changes to make the creation of scamp MEF catalogues and scamp header 
# files much more efficient:
# - The script was adapted to the new structure of scmapcat.py which
#   now can convert multiple images to scamp MEF files with a single
#   call. of the Python interpreter; before it was called one for
#   each input image!
# - Unnecessary splittings of the global ahead file are now avoided.
#   It is splitted once and the splittings are reused for each image;
#   before the same file splitting was done for each input image
#   separately!
#
# 14.04.2009:
# - I made the code more efficient in writing the scamp ahead files.
# - The script now has different processing modes for optimal astrometric
#   and relative photometric calibration.
#
# 05.02.2010:
# I introduced the an estimate of the maximum positional offset of
# sources as a comand line arguments. For WFI, the initial astrometry can be 
# off by up to two arcmin!
# 
# 31.03.2010:
# I fixed an error in the checking of the number of command line
# arguments - the case when no arguments were provided was not treated
# correctly.
#
# 17.09.2010:
# The mosaic type in case 'ahead' files are present is now 'SAME_CRVAL'
# and no longer 'FIX_FOCALPLANE' - recommendation by Emmanuel in the
# Astrometric scamp forum.
#
# 24.11.2010:
# We limit the maximum parallel channels for scamp to 16. Ludos new
# jade has considerably more and scamp refuses to process with a higher
# setting.
#
# 30.12.2010:
# We can now provide manually created standard star catalogues.
#
# 07.01.2011:
# Manual reference catalogues enter astrometric calibration with an
# astrometric weight of one now. Previously I set this to 1000 but
# this was only necessary because of wrong positional errors in some
# manually created standardstar catalogue.
#
# 27.05.2011:
# I corrected a bug that caused a non-unique identification of catalogue
# names (line CATNAME=`grep ...) for images with names 'image.fits' and
# 'ximage.fits' where 'image' is an arbitrary string. (Thanks to Mischa)

# File inclusions:
. ${INSTRUMENT:?}.ini

# define THELI_DEBUG and some other variables because of the '-u'
# script flag (the use of undefined variables will be treated as
# errors!)  THELI_DEBUG is used in the cleanTmpFiles function.
# 
THELI_DEBUG=${THELI_DEBUG:-""}
P_SCAMP=${P_SCAMP:-""}
P_ACLIENT=${P_ACLIENT:-""}
DIR=${DIR:-""}

##
## function definitions:
##
function printUsage 
{
  echo "SCRIPT NAME:"
  echo "    create_scamp_astrom_photom.sh"
  echo ""
  echo "ARGUMENTS:"
  echo "    1. main_dir.         1"
  echo "    2. set_dir.          1"
  echo "    3. image_extension   1"
  echo "    4. main_dir.         2"
  echo "    5. set_dir.          2"
  echo "    6. image_extension   2"
  echo "    ."
  echo "    ."
  echo "    n. (n-1) standardstar_catalog"
  echo "   (n.) MODE (OPTIONAL: astrom, photom or ALL;"
  echo "              default is ALL) "
  echo ""
  echo "    An arbitrary number of blocks 'main_dir.', 'set_dir.' and"
  echo "    'image_extension' can be provided."
  echo ""
  echo "    If a MODE is provided in the nth argument the script calls scamp"
  echo "    with optimised parameters for astrometry or relative"
  echo "    photometry. Omitting this argument (MODE ALL) runs scamp with"
  echo "    optimised parameters for astrometric calibration."
  echo ""
  echo "    ${SET} in the description below refers to the directory(ies)"
  echo "    /main dir./set dir./"
  echo ""
  echo "DESCRIPTION:"
  echo "    This script performs astrometric and relative photometric"
  echo "    calibration of THELI sets with Emanuel Bertins scamp program. The"
  echo "    script was tested against scamp V1.4.6 and scamp V1.7.0; The scamp"
  echo "    program itself is not contained in the THELI pipeline package and"
  echo "    needs, together with 'cdsclient' and the Python module 'pyfits',"
  echo "    to be installed separately (see the DEPNDENCY section below)"
  echo "    "
  echo "    The script substitutes the scripts"
  echo "    "
  echo "    - create_astrometrix_astrom(_run).sh"
  echo "    - create_photometrix.sh"
  echo "    "
  echo "    of the THELI ASTROMETRIX/PHOTOMETRIX processing. It ends with"
  echo "    header files in a \${SET}/headers_scamp_\${STANDARD}_\${MODE}"
  echo "    directory (the \${STANDARD} stands for the used standardstar"
  echo "    catalogue. In case of a manually created standardstar catalogue it"
  echo "    is 'MANUAL'. the MODE for the use processing mode; the MODE suffix"
  echo "    is omitted if the MODE is ALL).  The resulting header files can"
  echo "    then directly be used for image co-addition.  To use the scamp"
  echo "    headers the 'prepare_coadd_swarp.sh' script has to be called with"
  echo "    the '-eh headers_scamp_..' option. "
  echo ""
  echo "    If this script is run with MODES astrom and photom on some data set"
  echo "    the resulting headers can later be combined with the script "
  echo "    'merge_scamp_cats.sh'."
  echo "    "
  echo "    !! NOTE:"
  echo "       A call to 'create_absphotom_photometrix.sh' after this script"
  echo "       is still necessary to calculate the absolute zeropoint for a"
  echo "       coadded image "
  echo "    !!"
  echo ""
  echo "    The standardstar_catalog (n-1 th ot nth command line argument) can"
  echo "    be any astrometric catalogue supported by scamp. If a manually created"
  echo "    standardstar catalogue (in LDAC format) should be used, provide the"
  echo "    full, absolute path to that catalogue."
  echo ""
  echo "    Unless with Astrometrix it is possible to astrometrically and"
  echo "    photometrically calibrate multiple, associated sets"
  echo "    (e.g. multi-coour observations of the same target) simultaneously."
  echo "    The script performs its tasks and puts resulting files in the"
  echo "    directory ${SET}/astrom_photom_scamp (of the first given image"
  echo "    set) and subdirectories therein. From there, final headers are"
  echo "    moved to the appropriate directories at the end of processing."
  echo "         "
  echo "IMPORTANT DEPENDENCIES:"
  echo "    (1) The script calls the Python script 'scampcat.py' which creates"
  echo "        scamp MEF exposure catalogues from the individual THELI chip "
  echo "        catalogues. This Python script itself depends on pyfits."
  echo "    (2) scamp uses the configuration file 'scamp_astrom_photom.scamp'"
  echo "        which is set to define scamp astrometric contexts on the"
  echo "        basis of the FILTER and RUN keywords. "
  echo "    (3) The script checks whether a file ${INSTRUMENT}.ahead is present"
  echo "        in your reduction directory. If yes, it interprets it as a scamp"
  echo "        'additional' header containing 'focal plane' (first order "
  echo "        astrometric) information overriding corresponding entries in image"
  echo "        headers. In this case scamp is run with the "
  echo "        '-MOSAIC_TYPE SAME_CRVAL' option. We use '-MOSAIC_TYPE UNCHANGED' "
  echo "        otherwise."
  echo ""
  echo "TECHNICAL/IMPLEMENTATION INFORMATION:    "
  echo "    Most parts of the scripts concern transformations between"
  echo "    files that scamp needs/returns and the formats needed by"
  echo "    our THELI pipeline. Main differences and some more (random)"
  echo "    comments are:"
  echo "    "
  echo "    (1) Our THELI 'runs' are called 'astrometric contexts' in the scamp"
  echo "        language."
  echo "    (2) For each multi-chip exposure scamp expects ONE object catalogue"
  echo "        in MEF format. In THELI we work with individual catalogues"
  echo "        for each chip. If we would feed all individual chips into scamp"
  echo "        they are not recognised as being part of a specific exposure."
  echo "    (3) scamp outputs ONE header file with all chips for each exposure."
  echo "        In THELI we need one header per chip."
  echo "    (4) scamp only gives a final flux scale w.r.t. a specified zeropoint."
  echo "        We need the 'relative' magnitude offsets of each exposure w.r.t."
  echo "        a zero mean offset."
  echo "    (5) Absolute magnitudes are calculated in THELI outside the "
  echo "        astrometry/relative photometry step. For the time being we"
  echo "        do not provide the necessary interface (header keywords) to use "
  echo "        the absolute calibration within scamp."
  echo "    (6) Note that names for astrometric standardstar catalogues are slightly"
  echo "        different for scmap and THELI/ASTROMETRIX; "
  echo "        e.g. USNOB1 (THELI), but USNO-B1 (scamp)."
  echo "    (7) scamp crashes if images with a different number of chips "
  echo "        (same instrument) are found within one astrometric context. "
  echo "        Images with a different number of chips in different contexts "
  echo "        pose no problem. This can for instance happen because of "
  echo "        readout-problems for some chips in some exposures."
  echo ""
  echo "AUTHOR:"
  echo "    Thomas Erben (terben@astro.uni-bonn.de)"
  echo ""
}


function cleanTmpFiles
{
  if [ -z ${THELI_DEBUG} ]; then
    echo "Cleaning temporary files for script $0"
    test -f photdata.txt_$$             && rm photdata.txt_$$
    test -f photdata_relzp.txt_$$       && rm photdata_relzp.txt_$$
    test -f ${DIR}/catlist.txt_$$       && rm ${DIR}/catlist.txt_$$
    test -f ${DIR}/allscampfiles.txt_$$ && rm ${DIR}/allscampfiles.txt_$$

    i=1
    while [ ${i} -le ${NCHIPS} ]
    do
      test -f ${DIR}/${INSTRUMENT}_${i}.ahead_$$ && \
           rm ${DIR}/${INSTRUMENT}_${i}.ahead_$$

      i=$(( $i + 1 ))
    done
  else
    echo "Variable THELI_DEBUG set! No cleaning of temp. files in script $0"    
  fi
}

# Handling of program interruption by CRTL-C
trap "echo 'Script $0 interrupted!! Cleaning up and exiting!'; \
      cleanTmpFiles; exit 1" INT

##
## initial sanity checks
##
# check whether we have the external 'scamp' and 'aclient' programs at all:
if [ -z ${P_SCAMP} ] || [ -z ${P_ACLIENT} ] 
then
  echo "You need the external 'scamp' AND 'aclient' programs to"
  echo "use this script! The necessary variable(s) in"
  echo "your progs.ini seem(s) not to be set! Exiting!!"
  exit 1;
fi

# check validity of command line arguments:
# First a check on the number of the arguments:
if [ $# -lt 5 ] || \
  ([ $(( ($# - 2) % 3 )) -ne 0 ] && [ $(( ($# - 3) % 3 )) -ne 0 ]); then
  printUsage
  exit 1
fi

if [ $(( ($# - 2) % 3 )) -eq 0 ]; then
  # The number of different image directories we have to consider:
  NDIRS=$(( ($# - 2) / 3 ))

  # get the estimate for the maximum positional error of sources:
  POSMAXERRARG=$(( $# - 1 ))
  POSMAXERR=${!POSMAXERRARG}

  # get the used reference catalogue into a variable
  STARCAT=${!#}

  # the 'optimal' mode for the scamp run here is 'all':
  MODE=all

  # the SUFFIX for directory names:
  SUFFIX=${STARCAT}
else
  NDIRS=$(( ($# - 2) / 3 ))
  POSMAXERRARG=$(( $# - 2 ))
  POSMAXERR=${!POSMAXERRARG}
  STARCATARG=$(( $# - 1 )) 
  STARCAT=${!STARCATARG}

  MODE=${!#}

  if [ "${MODE}" != "all" ] && [ "${MODE}" != "astrom" ] && \
     [ "${MODE}" != "photom" ]; then
    printUsage
    exit 1
  fi
fi

# the SUFFIX for directory names. If we have a user provided astrometric
# standard star catalogue the sufix is 'MANUAL'. Otherwise it is the
# name of the standard star catalogue.
SUFFIX=${STARCAT}
if [ -f ${STARCAT} ]; then
  SUFFIX=MANUAL
fi

if [ "${MODE}" != "all" ]; then
  SUFFIX=${SUFFIX}_${MODE}
fi


# The location of the reduction directory; we need it to have
# a reference to the location of temporaray files:
DIR=`pwd`

# Test existence of image directory(ies) and create headers_scamp
# directories:
i=1 
j=2
k=1
while [ ${k} -le ${NDIRS} ]
do 
  if [ -d /${!i}/${!j} ]; then
    if [ -d /${!i}/${!j}/headers_scamp_${SUFFIX} ]; then
      echo "deleting old /${!i}/${!j}/headers_scamp_${SUFFIX}"
      rm -rf /${!i}/${!j}/headers_scamp_${SUFFIX}
    fi
    mkdir /${!i}/${!j}/headers_scamp_${SUFFIX}
  else
    echo "Can't find directory /${!i}/${!j}"; 
    exit 1;
  fi
  i=$(( ${i} + 3 ))
  j=$(( ${j} + 3 ))
  k=$(( ${k} + 1 ))
done

##
## Here the main script starts:
##

# all processing is performed in the 'first' image directory in
# a astrom_photom_scamp subdirectory:
cd /$1/$2/

if [ -d astrom_photom_scamp_${SUFFIX} ]; then
  echo "deleting old /$1/$2/astrom_photom_scamp_${SUFFIX}"
  rm -rf astrom_photom_scamp_${SUFFIX}
fi

mkdir -p astrom_photom_scamp_${SUFFIX}/cat
mkdir astrom_photom_scamp_${SUFFIX}/headers
mkdir astrom_photom_scamp_${SUFFIX}/plots

cd astrom_photom_scamp_${SUFFIX}/cat

# filter input catalogues to reject bad objects
i=1
j=2
k=3
l=1
NCATS=0

while [ ${l} -le ${NDIRS} ]
do 
  FILES=`${P_FIND} /${!i}/${!j}/cat/ -maxdepth 1 -name \*${!k}.cat`

  for CAT in ${FILES}
  do
    NCATS=$(( ${NCATS} + 1 ))

    BASE=`basename ${CAT} ${!k}.cat`
    # we filter away flagged objects except THOSE which are saturated!
    # we also require a minimum size (semi minor axis) of two pixels
    ${P_LDACFILTER} -i ${CAT} -t LDAC_OBJECTS \
         -c "((FLAGS<16)AND(B_IMAGE>2.0));" \
         -o ${BASE}${!k}.ldac

    # The follwowing two arrays are necessary to put headers
    # to the correct directories lateron.
    CATBASE[${NCATS}]=${BASE}
    CATDIR[${NCATS}]=/${!i}/${!j}
  done

  i=$(( ${i} + 3 ))
  j=$(( ${j} + 3 ))
  k=$(( ${k} + 3 ))
  l=$(( ${l} + 1 ))
done

# The aim of the next code part (up to the python call to S_SCAMPCAT
# is to create merged MEF catalogues for each exposures. In addition,
# scamp ahaed files need to be created. Most of the fiddling in the
# following is necessary because catalogues
# for individual chips might not be present (bad chips)

# First a simple list of all avilable individual catalogues.
# Its main purpose is to use it later to check whether certain catalogues
# are presnt or not. Doing this via 'grep' turned out to be significantly
# more efficient than bash file checks.
${P_FIND} . -name \*ldac | awk -F/ '{print $NF}' > ${DIR}/allscampfiles.txt_$$

# The following 'awk' construct cuts away everything after 
# the last '_' in the image names (including the underscore itself);
# we implicitely assume that the image extension DOES NOT contain '_'s. 
EXPOSURES=`cat ${DIR}/allscampfiles.txt_$$ |\
           ${P_GAWK} '{ n = split($1, a, "_"); 
                        name = ""; 
                        for(i = 1; i < (n-1); i++) {
                          name = name a[i] "_"
                        } 
                        name = name a[n-1]; 
                        print name;}' | sort | uniq`

# If present, split the instrument ahaed file in parts for every chip.
# We need to put together a separate one for each file (missing chips
# etc.)
AHEAD=0
if [ -f ${DIR}/${INSTRUMENT}.ahead ]; then
  ${P_GAWK} 'BEGIN {
               ext = "1"; 
               filename = "'${DIR}'/'${INSTRUMENT}'_" ext ".ahead_" '$$'; 
             } 
             { if($1 != "END") {
                 print $0 >> filename
               } else {
                 ext++;
                 filename = "'${DIR}'/'${INSTRUMENT}'_" ext ".ahead_" '$$'; 
               } 
             }' ${DIR}/${INSTRUMENT}.ahead
  AHEAD=1
fi

# The file catlist.txt_$$ is built for the python call to scampcat:
test -f ${DIR}/catlist.txt_$$ && rm ${DIR}/catlist.txt_$$

for EXPOSURE in ${EXPOSURES}
do
  # If an old scamp catalogue exists the python call for scampcat.py
  # (see below) would fail!
  test -f ./${EXPOSURE}_scamp.cat && rm -f ./${EXPOSURE}_scamp.cat

  i=1
  CATS=""
  MISSCHIP=0     # contains the missing chips in the form of a pasted
                 # string. If e. g. chips 19 and 25 are bad the variable
                 # would contain 01925 (read 0_19_25; '0' is always at
                 # the beginning)
  while [ ${i} -le ${NCHIPS} ]
  do
    # The following test for an image implicitely assumes that the
    # image ending does NOT start with a number: obvious but I mention
    # it just in case .... The '-x' option to grep ensures a proper
    # identification of images with names 'image.fits' and 'ximage.fits'
    # where 'image' and 'x' are arbitrary strings.
    # It is necessary as we allow for images with different endings in the 
    # image directories:
    CATNAME=`grep -x ''${EXPOSURE}'_'${i}'[^0-9].*\.ldac' \
             ${DIR}/allscampfiles.txt_$$`
  
    if [ -n "${CATNAME}" ]; then
      CATS="${CATS} ${CATNAME}"
    else
      MISSCHIP=${MISSCHIP}${i}
    fi
    i=$(( ${i} + 1 ))
  done
  echo "${CATS} ./${EXPOSURE}_scamp.cat" >> ${DIR}/catlist.txt_$$

  # Dummy external header containing focal plane and missing 
  # chip information.
  # They are used to distinguish different chip configurations
  # in an, otherwise, unique astrometric conetxt.
  test -f ./${EXPOSURE}_scamp.ahead && rm ./${EXPOSURE}_scamp.ahead

  i=1
  while [ ${i} -le ${NCHIPS} ]
  do
    CATNAME=`grep ''${EXPOSURE}'_'${i}'[^0-9].*\.ldac' \
             ${DIR}/allscampfiles.txt_$$`

    if [ -n "${CATNAME}" -a -f "${CATNAME}" ]; then
      if [ ${AHEAD} -eq 1 ]; then
        cat  ${DIR}/${INSTRUMENT}_${i}.ahead_$$ 
      fi
      echo "MISSCHIP= '${MISSCHIP}'" 
      echo "END      "               
    fi
    i=$(( ${i} + 1 ))
  done > ./${EXPOSURE}_scamp.ahead
done

${P_PYTHON} ${S_SCAMPCAT} ${DIR}/catlist.txt_$$

# now call scamp:
cd ../headers

# scamp mosaic type dependent on the availability of a first order astrometric
# solution:
if [ ${AHEAD} -eq 1 ]; then
  MOSAICTYPE="-MOSAIC_TYPE SAME_CRVAL"
else
  MOSAICTYPE="-MOSAIC_TYPE UNCHANGED"
fi

# for some reason scamp gives better results if the
# 'ordering' of files is shuffled in time! Do not ask!
if [ "${MODE}" != "photom" ]; then
  FILES=`${P_FIND} ../cat/ -name \*scamp.cat |\
         awk '{line[NR] = $0} END { srand();
           for (i = 1; i <= NR; i++) { 
             n = int(rand() * NR) + 1; 
             temp = line[i]; 
             line[i] = line[n];
             line[n] = temp;
           }
           for (i = 1; i <= NR; i++) {
             print line[i];
           }}'`
  # For astrometric calibration we can
  # consider saturated sources:
  FLAGS="-FLAGS_MASK 240"
else
  FILES=`${P_FIND} ../cat/ -name \*scamp.cat | ${P_SORT}`
  # For photometric calibration we do not want to
  # consider saturated sources:
  FLAGS="-FLAGS_MASK 244"
fi
 
# scamp (at least up to V1.7) has an upper limit on the parallel
# channels of 16.  With Ludos new jade we reached this:
if [ ${NPARA} -gt 16 ]; then
  NPARA=16
fi

ASTREF=""
if [ -f ${STARCAT} ]; then
  ASTREF="-ASTREF_CATALOG FILE -ASTREFCAT_NAME ${STARCAT}"
else
  ASTREF="-ASTREF_CATALOG ${STARCAT}"
fi 

${P_SCAMP} ${FILES} \
           -c ${CONF}/scamp_astrom_photom.scamp \
           -CDSCLIENT_EXEC ${P_ACLIENT} \
           -POSITION_MAXERR ${POSMAXERR} \
           -NTHREADS ${NPARA} \
           ${ASTREF} ${FLAGS} ${MOSAICTYPE}

if [ $? -ne 0 ]
then
  echo "scamp call failed!! Exiting!!"
  cleanTmpFiles
  exit 1
fi

# scamp creates the headers in the directory where the catalogs
# are:
${P_FIND}  ../cat/ -name \*.head -exec mv {} . \;

# we want the diagnostic plots in an own directory:
mv fgroups*         ../plots
mv distort*         ../plots
mv astr_interror2d* ../plots
mv astr_interror1d* ../plots
mv astr_referror2d* ../plots
mv astr_referror1d* ../plots
mv astr_chi2*       ../plots
mv psphot_error*    ../plots

# now get the relative magnitude offsets from the FLXSCALES
# estimated by scamp:
test -f photdata.txt_$$ && rm photdata.txt_$$

# Because the flux scales refer to an image normalised to one
# second we need to obtain the exposure times of all frames
# first. We also get the SCAMP flux scale and the photometric 
# instrument:
for EXPOSURE in ${EXPOSURES}
do
  EXPTIME=`${P_LDACTOASC} -i ../cat/${EXPOSURE}_scamp.cat \
                     -t LDAC_IMHEAD -s |\
           fold | grep EXPTIME | ${P_GAWK} '{print $3}'`
  FLXSCALE=`grep FLXSCALE ${EXPOSURE}_scamp.head | uniq |\
            ${P_GAWK} '{print $2}'`
  PHOTINST=`grep PHOTINST ${EXPOSURE}_scamp.head | uniq |\
            ${P_GAWK} '{print $2}'`

  echo ${EXPOSURE}" "${EXPTIME}" "${FLXSCALE}" "${PHOTINST} >> photdata.txt_$$
done

# The following 'awk' script calculates relative zeropoints 
# and THELI fluxscales for the different photometric contexts: 
${P_GAWK} 'BEGIN {maxphotinst = 1;}
           { exposure[NR]       = $1; 
             exptime[NR]        = $2; 
             flxscale_scamp[NR] = $3;
             photinst[NR]       = $4
             val[NR]            = -2.5 * log($3 * $2) / log(10); 
             m[$4]              = m[$4] + val[NR]
             nphotinst[$4]      = nphotinst[$4] + 1 
             if ($4 > maxphotinst) {maxphotinst = $4}} 
           END {
             for (i = 1; i <= maxphotinst; i++) {  
               m[i] = m[i] / nphotinst[i];
             } 
             for (i = 1; i <= NR; i++) {
               relzp[i] = val[i] - m[photinst[i]];   
               flxscale_theli[i] = (10**(-0.4 * relzp[i])) / exptime[i];
               printf("%s %f %e\n", 
                 exposure[i], relzp[i], flxscale_theli[i]);  
             }
           }' photdata.txt_$$ > photdata_relzp.txt_$$

# now split the exposure catalogues for the indivudual chips
# and add the RZP and FLXSCALE header keywords. Put the headers 
# into appropriate headers_scamp directories
#
while read EXPOSURE RELZP FLXSCALE
do
  i=1
  j=1  # counts the actually available chips!
  while [ ${i} -le ${NCHIPS} ]
  do
    # we need to take care of catalogs that may not be
    # present (bad chips)!
    CATNAME=`grep ''${EXPOSURE}'_'${i}'[^0-9].*\.ldac' \
             ${DIR}/allscampfiles.txt_$$`
    if [ -n "${CATNAME}" -a -f ../cat/${CATNAME} ]
    then
      # first rename the SCAMP header keyword FLXSCALE
      # to FLSCALE. We need FLXSCALE for the THELI
      # flux scaling later:
      sed -e 's/FLXSCALE/FLSCALE /' ${EXPOSURE}_scamp.head |\
      ${P_GAWK} 'BEGIN {ext = '${j}'; nend = 0} 
                 {
                   if(nend < ext) {
                     if($1 == "END") {
                       nend++; 
                       next; 
                     } 
                     if(nend == (ext-1)) { print $0 } 
                   } 
                 }
                 END { printf("RZP     = %20f / THELI relative zeropoint\n", 
                              '${RELZP}');
                       printf("FLXSCALE= %20E / THELI relative flux scale\n", 
                              '${FLXSCALE}');
                       printf("END\n")
                 }' > ${EXPOSURE}_$i.head
      j=$(( $j + 1 ))
    fi
    i=$(( $i + 1 ))
  done
done < photdata_relzp.txt_$$

i=1
while [ ${i} -le ${NCATS} ]
do
  if [ -f ${CATBASE[$i]}.head ]; then
    mv ${CATBASE[$i]}*head ${CATDIR[$i]}/headers_scamp_${SUFFIX}
  fi
  
  i=$(( ${i} + 1 )) 
done

# clean up temporary files and bye
cleanTmpFiles

cd ${DIR}

exit 0
