#!/bin/bash -u

# create catalogues suitable for absolute photometric calibration
# with 'create_abs_photo_info.sh'. The script takes astrometric
# information from a scamp processing and calculates 'correct' sky
# coordinates from objects catalogues (extracted from standard star
# fields). Afterwards the standard star object catalogues are
# merged with a reference standard star sources (e.g. from Landolt,
# Stetson, Sloan ....)

#$1: main directory
#$2: science dir. (the catalogues are assumed to be in $1/$2/cat)
#$3: image extension
#$4: astrometry standardstar catalogue used for the scamp calibration
#    (the script needs the scamp headers which are assumed to be in 
#    $1/$2/headers_scamp_$4)
#$5: photometric standard star catalogue with full path

. ${INSTRUMENT:?}.ini 

# give meaningful names to command line arguments:
MD=$1
SD=$2
EXTENSION=$3
STANDARDCAT=$4
PHOTCAT=$5

for CHIP in ${!#}
do
  CATS=`find ${MD}/${SD}/cat -name \*_${CHIP}${EXTENSION}.cat`

  for CAT in ${CATS}
  do
    BASE=`basename ${CAT} .cat`
    HEADBASE=`basename ${CAT} ${EXTENSION}.cat`
    ${P_LDACCONV} -i ${CAT} -o ${MD}/${SD}/cat/${BASE}_ldac.cat \
                  -b ${CHIP} -c ${INSTRUMENT} -f dummy 

    # get 'correct' astrometric scamp information to the catalogues:    
    ${S_APLASTROMSCAMP} -i ${MD}/${SD}/cat/${BASE}_ldac.cat \
         -o ${MD}/${SD}/cat/${BASE}_ldac_corr.cat \
         -s ${MD}/${SD}/headers_scamp_${STANDARDCAT}/${HEADBASE}.head \
         -t OBJECTS -p Xpos Ypos -r Ra Dec

    ${P_MAKEJOIN} -i /${MD}/${SD}/cat/${BASE}_ldac_corr.cat \
                  -o ${TEMPDIR}/tmp.cat0_$$ \
		  -c ${DATACONF}/stdphotom_prepare_make_join.conf
    # calculate object magnitudes with a magzeropoint of 0 and
    # an exposure time normalisation of 1s (1.08574 is 2.5 / log(10)):
    ${P_LDACCALC} -i ${TEMPDIR}/tmp.cat0_$$ \
                  -o ${TEMPDIR}/tmp.cat00_$$ \
                  -t OBJECTS -c "(-1.08574*log(FLUX_AUTO/EXPTIME));" \
                  -n MAG_AUTO_corr "exposure time corr. MAG_AUTO" -k FLOAT
    ${P_LDACRENTAB} -i ${TEMPDIR}/tmp.cat00_$$\
                    -o ${TEMPDIR}/tmp.cat1_$$ -t OBJECTS STDTAB
    ${P_LDACRENKEY} -i ${TEMPDIR}/tmp.cat1_$$ -o ${TEMPDIR}/tmp.cat2_$$ \
                    -t STDTAB -k A_WORLD A_WCS B_WORLD B_WCS \
                                 THETA_J2000 THETAWCS
    ${P_ASSOCIATE} -i ${TEMPDIR}/tmp.cat2_$$ ${PHOTCAT}\
		   -o ${TEMPDIR}/tmp.cat3_$$ ${TEMPDIR}/tmp.cat4_$$ \
                   -t STDTAB -c ${DATACONF}/stdphotom_prepare_associate.conf

    ${P_LDACFILTER} -i ${TEMPDIR}/tmp.cat3_$$ -o ${TEMPDIR}/tmp.cat5_$$ \
                    -c "(Pair_1>0);" -t STDTAB

    if [ "$?" -eq "0" ]; then
      ${P_LDACFILTER} -i ${TEMPDIR}/tmp.cat4_$$ \
                      -o ${TEMPDIR}/tmp.cat6_$$ -c "(Pair_0>0);" -t STDTAB
      ${P_ASSOCIATE} -i ${TEMPDIR}/tmp.cat5_$$ ${TEMPDIR}/tmp.cat6_$$ \
                     -o ${TEMPDIR}/tmp.cat7_$$ ${TEMPDIR}/tmp.cat8_$$ \
      	       -t STDTAB -c ${DATACONF}/stdphotom_prepare_associate.conf
      ${P_MAKESSC} -i ${TEMPDIR}/tmp.cat7_$$ ${TEMPDIR}/tmp.cat8_$$ \
         	   -o /${MD}/${SD}/cat/${BASE}_merg.cat\
        	   -t STDTAB \
                   -c ${DATACONF}/stdphotom_prepare_make_ssc.conf
    fi
  done
  ${P_LDACPASTE} -i /${MD}/${SD}/cat/*_${CHIP}${EXTENSION}_merg.cat \
                 -o /${MD}/${SD}/cat/chip_${CHIP}_merg.cat -t PSSC
done

# clean up
rm  ${TEMPDIR}/tmp.cat*_$$
