#!/bin/bash 

# the script merges info from the singleastrom
# and the check_science_PSF step. It transfers
# e1, e2 from the KSB cats to the SExtractor cats
# giving the "new" catalogs the same .cat as the old
# ones. Hence these catalogs can be used in the
# further processing having KSB info. Also this step
# is not absolutely necssary in the processing and
# can be omitted if this PSF information is not 
# important.
#
# 30.05.04:
# temporary files go to a TEMPDIR directory 
#
# 17.01.2007:
# - If KSB catalogues for certain files are not present
#   they will be filled up with dummy values for the four
#   keys e1, e2, rh and cl (e1=e2=-2.0; rh=0.0; cl=0).
#   We no longer quit the script
#   if KSB information is not present for all files.
#   In this way we still can use the present lensing
#   information.
# - filenames for temporary files are unique now including
#   the process number.
#
# 07.02.2007:
# I substituted two 'ls' commands with 'find' equivalents 
# to avoid 'argument list too long' errors.
#
# 28.07.2007:
# I corrected a bug when seeting the BASE variable in a loop.
#
# 06.05.2008: 
# I now copy catalogues always; before they were only copied if the
# 'copy' subdiretory (below 'cat') did not yet exist.  It could then
# happen that a faulty previous run of this script led to an empty
# 'copy' directory and to no transfer of KSB info!
#
# 05.10.2008:
# I make sure that temporary files are removed at the end of
# the script.
#
# 16.07.2009:
# The name of the catalogue subdirectory now has to be provided as
# command line argument.
#
# 23.10.2009:
# bug fix in the addition of dummy lensing keywords (when appropriate
# values could not be determined)

#$1: main dir.
#$2: science dir.
#$3: image extension (ext) on ..._iext.fits (i is the chip number)
#$4: catalogue directory

. ${INSTRUMENT:?}.ini

# first make a security copy of catalogues:
test -d /$1/$2/$4/copy || mkdir /$1/$2/$4/copy
${P_FIND} /$1/$2/$4/ -maxdepth 1 -name \*$3.cat -exec cp {} /$1/$2/$4/copy \;

# now transfer KSB info
${P_FIND} /$1/$2/$4/copy/ -name \*$3.cat > ${TEMPDIR}/mergecats_$$

# test if all KSB cats are there; write missing ones to a file!
test -f ${TEMPDIR}/fail_$$ && rm ${TEMPDIR}/fail_$$

while read CAT
do
  BASE=`basename ${CAT} .cat`
  if [ ! -f "/$1/$2/$4/${BASE}_ksb_tmp.cat2" ]; then
    echo "/$1/$2/$4/${BASE}_ksb_tmp.cat2 is not there" >> fail_$$
  fi
done < ${TEMPDIR}/mergecats_$$  

while read CAT
do
  # The following ldacfilter is necessary because a 
  # corresponding cut was applied when running the KSB
  # algorithm (TO BE CHANGED because we would like to
  # keep all objects!)
  ${P_LDACFILTER} -i ${CAT} -o ${TEMPDIR}/tmp1_$$.cat\
                  -t LDAC_OBJECTS\
                  -c "(FLUX_RADIUS>0.0)AND(FLUX_RADIUS<10.0);"
  BASE=`basename ${CAT} .cat`
  if [ -f "/$1/$2/$4/${BASE}_ksb_tmp.cat2" ]; then
    ${P_LDACRENTAB} -i /$1/$2/$4/${BASE}_ksb_tmp.cat2 \
     	            -o ${TEMPDIR}/tmp_$$.cat \
                    -t OBJECTS LDAC_OBJECTS
    ${P_LDACJOINKEY} -i ${TEMPDIR}/tmp1_$$.cat \
                     -o /$1/$2/$4/${BASE}.cat \
	             -p ${TEMPDIR}/tmp_$$.cat \
	             -t LDAC_OBJECTS \
                     -k e1 e2 rh cl snratio  
    rm ${TEMPDIR}/tmp_$$.cat ${TEMPDIR}/tmp1_$$.cat		     
  else
    # if we have no lensing information we add dummy
    # information to have consistent catalogues
    ${P_LDACADDKEY} -i ${TEMPDIR}/tmp1_$$.cat \
                    -o /$1/$2/$4/${BASE}.cat \
                    -t LDAC_OBJECTS \
                    -k e1 -2.0 FLOAT "dummy value" \
                       e2 -2.0 FLOAT "dummy value" \
                       rh 0.0 FLOAT "dummy value" \
                       cl 0 SHORT "dummy value" \
                       snratio 0 FLOAT "dummy value"
    rm ${TEMPDIR}/tmp1_$$.cat
  fi
done < ${TEMPDIR}/mergecats_$$ 

test -f ${TEMPDIR}/mergecats_$$ && rm ${TEMPDIR}/mergecats_$$
test -f ${TEMPDIR}/tmp_$$.cat   && rm ${TEMPDIR}/tmp_$$.cat
test -f ${TEMPDIR}/tmp1_$$.cat  && rm ${TEMPDIR}/tmp1_$$.cat


